(function() {
  var DatabaseStore, Emitter, NylasLongConnection, url, _,
    __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; };

  Emitter = require('event-kit').Emitter;

  DatabaseStore = require('nylas-exports').DatabaseStore;

  url = require('url');

  _ = require('underscore');

  NylasLongConnection = (function() {
    NylasLongConnection.Status = {
      None: 'none',
      Connecting: 'connecting',
      Connected: 'connected',
      Closed: 'closed',
      Ended: 'ended'
    };

    function NylasLongConnection(api, accountId, config) {
      this.closeIfDataStops = __bind(this.closeIfDataStops, this);
      this.onProcessBuffer = __bind(this.onProcessBuffer, this);
      this._api = api;
      this._accountId = accountId;
      this._config = config;
      this._emitter = new Emitter;
      this._status = NylasLongConnection.Status.None;
      this._req = null;
      this._pingTimeout = null;
      this._buffer = null;
      this._deltas = [];
      this._flushDeltasDebounced = _.debounce((function(_this) {
        return function() {
          var last;
          if (_this._deltas.length === 0) {
            return;
          }
          last = _this._deltas[_this._deltas.length - 1];
          _this._emitter.emit('deltas-stopped-arriving', _this._deltas);
          _this._config.setCursor(last.cursor);
          return _this._deltas = [];
        };
      })(this), 1000);
      this;
    }

    NylasLongConnection.prototype.accountId = function() {
      return this._accountId;
    };

    NylasLongConnection.prototype.hasCursor = function() {
      return !!this._config.getCursor();
    };

    NylasLongConnection.prototype.withCursor = function(callback) {
      var cursor;
      cursor = this._config.getCursor();
      if (cursor) {
        return callback(cursor);
      }
      return this._api.makeRequest({
        path: "/delta/latest_cursor",
        accountId: this._accountId,
        method: 'POST',
        success: (function(_this) {
          return function(_arg) {
            var cursor;
            cursor = _arg.cursor;
            console.log("Obtained stream cursor " + cursor + ".");
            _this._config.setCursor(cursor);
            return callback(cursor);
          };
        })(this)
      });
    };

    NylasLongConnection.prototype.status = function() {
      return this.status;
    };

    NylasLongConnection.prototype.setStatus = function(status) {
      if (this._status === status) {
        return;
      }
      this._status = status;
      return this._config.setStatus(status);
    };

    NylasLongConnection.prototype.onDeltas = function(callback) {
      return this._emitter.on('deltas-stopped-arriving', callback);
    };

    NylasLongConnection.prototype.onProcessBuffer = function() {
      var bufferJSONs, delta, deltaJSON, e, _i, _len, _results;
      bufferJSONs = this._buffer.split('\n');
      this._buffer = bufferJSONs.pop();
      _results = [];
      for (_i = 0, _len = bufferJSONs.length; _i < _len; _i++) {
        deltaJSON = bufferJSONs[_i];
        if (deltaJSON.length === 0) {
          continue;
        }
        delta = null;
        try {
          delta = JSON.parse(deltaJSON);
        } catch (_error) {
          e = _error;
          console.log("" + deltaJSON + " could not be parsed as JSON.", e);
        }
        if (delta) {
          if (!delta.cursor) {
            throw new Error('Received delta with no cursor!');
          }
          this._deltas.push(delta);
          _results.push(this._flushDeltasDebounced());
        } else {
          _results.push(void 0);
        }
      }
      return _results;
    };

    NylasLongConnection.prototype.start = function() {
      var token, _ref;
      if (!this._config.ready()) {
        return;
      }
      if ((_ref = this._status) !== NylasLongConnection.Status.None && _ref !== NylasLongConnection.Status.Closed) {
        return;
      }
      token = this._api.accessTokenForAccountId(this._accountId);
      if (token == null) {
        return;
      }
      if (this._req) {
        return;
      }
      return this.withCursor((function(_this) {
        return function(cursor) {
          var lib, options;
          if (_this.status === NylasLongConnection.Status.Ended) {
            return;
          }
          options = url.parse("" + _this._api.APIRoot + "/delta/streaming?cursor=" + cursor + "&exclude_folders=false&exclude_metadata=false&exclude_account=false");
          options.auth = "" + token + ":";
          if (_this._api.APIRoot.indexOf('https') === -1) {
            lib = require('http');
          } else {
            options.port = 443;
            lib = require('https');
          }
          _this._req = lib.request(options, function(res) {
            var processBufferThrottled;
            if (res.statusCode !== 200) {
              res.on('data', function(chunk) {
                var error;
                if (chunk.toString().indexOf('Invalid cursor') > 0) {
                  error = new Error('Delta Connection: Cursor is invalid. Need to blow away local cache.');
                  NylasEnv.config.unset("nylas." + _this._accountId + ".cursor");
                  DatabaseStore._handleSetupError(error);
                }
                return _this.close();
              });
              return;
            }
            _this._buffer = '';
            processBufferThrottled = _.throttle(_this.onProcessBuffer, 400, {
              leading: false
            });
            res.setEncoding('utf8');
            res.on('close', function() {
              return _this.close();
            });
            return res.on('data', function(chunk) {
              _this.closeIfDataStops();
              if (chunk === '\n' && (_this._buffer.length === 0 || _this._buffer[-1] === '\n')) {
                return;
              }
              _this._buffer += chunk;
              return processBufferThrottled();
            });
          });
          _this._req.setTimeout(60 * 60 * 1000);
          _this._req.setSocketKeepAlive(true);
          _this._req.on('error', function() {
            return _this.close();
          });
          _this._req.on('socket', function(socket) {
            _this.setStatus(NylasLongConnection.Status.Connecting);
            return socket.on('connect', function() {
              _this.setStatus(NylasLongConnection.Status.Connected);
              return _this.closeIfDataStops();
            });
          });
          return _this._req.end();
        };
      })(this));
    };

    NylasLongConnection.prototype.close = function() {
      if (this._status === NylasLongConnection.Status.Closed) {
        return;
      }
      this.setStatus(NylasLongConnection.Status.Closed);
      return this.cleanup();
    };

    NylasLongConnection.prototype.closeIfDataStops = function() {
      if (this._pingTimeout) {
        clearTimeout(this._pingTimeout);
      }
      return this._pingTimeout = setTimeout((function(_this) {
        return function() {
          _this._pingTimeout = null;
          return _this.close();
        };
      })(this), 15 * 1000);
    };

    NylasLongConnection.prototype.end = function() {
      this.setStatus(NylasLongConnection.Status.Ended);
      return this.cleanup();
    };

    NylasLongConnection.prototype.cleanup = function() {
      if (this._pingTimeout) {
        clearInterval(this._pingTimeout);
      }
      this._pingTimeout = null;
      this._buffer = '';
      if (this._req) {
        this._req.end();
        this._req.abort();
        return this._req = null;
      }
    };

    return NylasLongConnection;

  })();

  module.exports = NylasLongConnection;

}).call(this);
