(function() {
  var Account, Actions, DatabaseStore, DatabaseTransaction, NylasLongConnection, NylasSyncWorker, Thread, _, _ref;

  _ = require('underscore');

  _ref = require('nylas-exports'), Actions = _ref.Actions, DatabaseStore = _ref.DatabaseStore, DatabaseTransaction = _ref.DatabaseTransaction, Account = _ref.Account, Thread = _ref.Thread;

  NylasLongConnection = require('../lib/nylas-long-connection');

  NylasSyncWorker = require('../lib/nylas-sync-worker');

  describe("NylasSyncWorker", function() {
    beforeEach(function() {
      this.apiRequests = [];
      this.api = {
        APIRoot: 'https://api.nylas.com',
        pluginsSupported: true,
        accessTokenForAccountId: (function(_this) {
          return function() {
            return '123';
          };
        })(this),
        makeRequest: (function(_this) {
          return function(requestOptions) {
            return _this.apiRequests.push({
              requestOptions: requestOptions
            });
          };
        })(this),
        getCollection: (function(_this) {
          return function(account, model, params, requestOptions) {
            return _this.apiRequests.push({
              account: account,
              model: model,
              params: params,
              requestOptions: requestOptions
            });
          };
        })(this),
        getThreads: (function(_this) {
          return function(account, params, requestOptions) {
            return _this.apiRequests.push({
              account: account,
              model: 'threads',
              params: params,
              requestOptions: requestOptions
            });
          };
        })(this)
      };
      this.apiCursorStub = void 0;
      spyOn(NylasSyncWorker.prototype, 'fetchAllMetadata').andCallFake(function(cb) {
        return cb();
      });
      spyOn(DatabaseTransaction.prototype, 'persistJSONBlob').andReturn(Promise.resolve());
      spyOn(DatabaseStore, 'findJSONBlob').andCallFake((function(_this) {
        return function(key) {
          if (key === ("NylasSyncWorker:" + TEST_ACCOUNT_ID)) {
            return Promise.resolve(_.extend({}, {
              "cursor": _this.apiCursorStub,
              "contacts": {
                busy: true,
                complete: false
              },
              "calendars": {
                busy: false,
                complete: true
              }
            }));
          } else if (key.indexOf('ContactRankings') === 0) {
            return Promise.resolve([]);
          } else {
            throw new Error("Not stubbed! " + key);
          }
        };
      })(this));
      this.account = new Account({
        clientId: TEST_ACCOUNT_CLIENT_ID,
        serverId: TEST_ACCOUNT_ID,
        organizationUnit: 'label'
      });
      this.worker = new NylasSyncWorker(this.api, this.account);
      this.worker._metadata = {
        "a": [
          {
            "id": "b"
          }
        ]
      };
      this.connection = this.worker.connection();
      spyOn(this.connection, 'start');
      return advanceClock();
    });
    it("should reset `busy` to false when reading state from disk", function() {
      this.worker = new NylasSyncWorker(this.api, this.account);
      spyOn(this.worker, 'resume');
      advanceClock();
      return expect(this.worker.state().contacts.busy).toEqual(false);
    });
    describe("start", function() {
      it("should open the delta connection", function() {
        this.worker.start();
        advanceClock();
        return expect(this.connection.start).toHaveBeenCalled();
      });
      it("should start querying for model collections and counts that haven't been fully cached", function() {
        var countsRequested, modelsRequested;
        this.worker.start();
        advanceClock();
        expect(this.apiRequests.length).toBe(12);
        modelsRequested = _.compact(_.map(this.apiRequests, function(_arg) {
          var model;
          model = _arg.model;
          return model;
        }));
        expect(modelsRequested).toEqual(['threads', 'messages', 'labels', 'drafts', 'contacts', 'events']);
        countsRequested = _.compact(_.map(this.apiRequests, function(_arg) {
          var requestOptions, _ref1;
          requestOptions = _arg.requestOptions;
          if (((_ref1 = requestOptions.qs) != null ? _ref1.view : void 0) === 'count') {
            return requestOptions.path;
          }
        }));
        expect(modelsRequested).toEqual(['threads', 'messages', 'labels', 'drafts', 'contacts', 'events']);
        return expect(countsRequested).toEqual(['/threads', '/messages', '/labels', '/drafts', '/contacts', '/events']);
      });
      it("should fetch 1000 labels and folders, to prevent issues where Inbox is not in the first page", function() {
        var labelsRequest;
        labelsRequest = _.find(this.apiRequests, function(r) {
          return r.model === 'labels';
        });
        return expect(labelsRequest.params.limit).toBe(1000);
      });
      it("should mark incomplete collections as `busy`", function() {
        var collection, nextState, _i, _len, _ref1, _results;
        this.worker.start();
        advanceClock();
        nextState = this.worker.state();
        _ref1 = ['contacts', 'threads', 'drafts', 'labels'];
        _results = [];
        for (_i = 0, _len = _ref1.length; _i < _len; _i++) {
          collection = _ref1[_i];
          _results.push(expect(nextState[collection].busy).toEqual(true));
        }
        return _results;
      });
      it("should initialize count and fetched to 0", function() {
        var collection, nextState, _i, _len, _ref1, _results;
        this.worker.start();
        advanceClock();
        nextState = this.worker.state();
        _ref1 = ['contacts', 'threads', 'drafts', 'labels'];
        _results = [];
        for (_i = 0, _len = _ref1.length; _i < _len; _i++) {
          collection = _ref1[_i];
          expect(nextState[collection].fetched).toEqual(0);
          _results.push(expect(nextState[collection].count).toEqual(0));
        }
        return _results;
      });
      it("after failures, it should attempt to resume periodically but back off as failures continue", function() {
        var simulateNetworkFailure;
        simulateNetworkFailure = (function(_this) {
          return function() {
            _this.apiRequests[1].requestOptions.error({
              statusCode: 400
            });
            return _this.apiRequests = [];
          };
        })(this);
        spyOn(this.worker, 'resume').andCallThrough();
        this.worker.start();
        expect(this.worker.resume.callCount).toBe(1);
        simulateNetworkFailure();
        expect(this.worker.resume.callCount).toBe(1);
        advanceClock(4000);
        expect(this.worker.resume.callCount).toBe(2);
        simulateNetworkFailure();
        expect(this.worker.resume.callCount).toBe(2);
        advanceClock(4000);
        expect(this.worker.resume.callCount).toBe(2);
        advanceClock(4000);
        expect(this.worker.resume.callCount).toBe(3);
        simulateNetworkFailure();
        expect(this.worker.resume.callCount).toBe(3);
        advanceClock(4000);
        expect(this.worker.resume.callCount).toBe(3);
        advanceClock(4000);
        expect(this.worker.resume.callCount).toBe(3);
        advanceClock(4000);
        expect(this.worker.resume.callCount).toBe(4);
        simulateNetworkFailure();
        expect(this.worker.resume.callCount).toBe(4);
        advanceClock(4000);
        expect(this.worker.resume.callCount).toBe(4);
        advanceClock(4000);
        expect(this.worker.resume.callCount).toBe(4);
        advanceClock(4000);
        expect(this.worker.resume.callCount).toBe(4);
        advanceClock(4000);
        expect(this.worker.resume.callCount).toBe(4);
        advanceClock(4000);
        return expect(this.worker.resume.callCount).toBe(5);
      });
      it("handles the request as a failure if we try and grab labels or folders without an 'inbox'", function() {
        var request;
        spyOn(this.worker, 'resume').andCallThrough();
        this.worker.start();
        expect(this.worker.resume.callCount).toBe(1);
        request = _.findWhere(this.apiRequests, {
          model: 'labels'
        });
        request.requestOptions.success([]);
        expect(this.worker.resume.callCount).toBe(1);
        advanceClock(30000);
        return expect(this.worker.resume.callCount).toBe(2);
      });
      return it("handles the request as a success if we try and grab labels or folders and it includes the 'inbox'", function() {
        var request;
        spyOn(this.worker, 'resume').andCallThrough();
        this.worker.start();
        expect(this.worker.resume.callCount).toBe(1);
        request = _.findWhere(this.apiRequests, {
          model: 'labels'
        });
        request.requestOptions.success([
          {
            name: "inbox"
          }, {
            name: "archive"
          }
        ]);
        expect(this.worker.resume.callCount).toBe(1);
        advanceClock(30000);
        return expect(this.worker.resume.callCount).toBe(1);
      });
    });
    describe("delta streaming cursor", function() {
      return it("should read the cursor from the database, and the old config format", function() {
        var connection, worker;
        spyOn(NylasLongConnection.prototype, 'withCursor').andCallFake((function(_this) {
          return function() {};
        })(this));
        this.apiCursorStub = void 0;
        worker = new NylasSyncWorker(this.api, this.account);
        connection = worker.connection();
        expect(connection.hasCursor()).toBe(false);
        advanceClock();
        expect(connection.hasCursor()).toBe(false);
        spyOn(NylasEnv.config, 'get').andCallFake((function(_this) {
          return function(key) {
            if (key === ("nylas." + _this.account.id + ".cursor")) {
              return 'old-school';
            }
            return void 0;
          };
        })(this));
        worker = new NylasSyncWorker(this.api, this.account);
        connection = worker.connection();
        advanceClock();
        expect(connection.hasCursor()).toBe(true);
        expect(connection._config.getCursor()).toEqual('old-school');
        this.apiCursorStub = "new-school";
        worker = new NylasSyncWorker(this.api, this.account);
        connection = worker.connection();
        expect(connection.hasCursor()).toBe(false);
        advanceClock();
        expect(connection.hasCursor()).toBe(true);
        return expect(connection._config.getCursor()).toEqual('new-school');
      });
    });
    describe("when a count request completes", function() {
      beforeEach(function() {
        this.worker.start();
        advanceClock();
        this.request = this.apiRequests[0];
        return this.apiRequests = [];
      });
      return it("should update the count on the collection", function() {
        var nextState;
        this.request.requestOptions.success({
          count: 1001
        });
        nextState = this.worker.state();
        return expect(nextState.threads.count).toEqual(1001);
      });
    });
    describe("resume", function() {
      it("should fetch metadata first and fetch other collections when metadata is ready", function() {
        var fetchAllMetadataCallback;
        fetchAllMetadataCallback = null;
        jasmine.unspy(NylasSyncWorker.prototype, 'fetchAllMetadata');
        spyOn(NylasSyncWorker.prototype, 'fetchAllMetadata').andCallFake((function(_this) {
          return function(cb) {
            return fetchAllMetadataCallback = cb;
          };
        })(this));
        spyOn(this.worker, 'fetchCollection');
        this.worker._state = {};
        this.worker.resume();
        expect(this.worker.fetchAllMetadata).toHaveBeenCalled();
        expect(this.worker.fetchCollection.calls.length).toBe(0);
        fetchAllMetadataCallback();
        return expect(this.worker.fetchCollection.calls.length).not.toBe(0);
      });
      it("should not fetch metadata pages if pluginsSupported is false", function() {
        this.api.pluginsSupported = false;
        spyOn(NylasSyncWorker.prototype, '_fetchWithErrorHandling');
        spyOn(this.worker, 'fetchCollection');
        this.worker._state = {};
        this.worker.resume();
        expect(this.worker._fetchWithErrorHandling).not.toHaveBeenCalled();
        return expect(this.worker.fetchCollection.calls.length).not.toBe(0);
      });
      it("should fetch collections for which `shouldFetchCollection` returns true", function() {
        spyOn(this.worker, 'fetchCollection');
        spyOn(this.worker, 'shouldFetchCollection').andCallFake((function(_this) {
          return function(collection) {
            return collection === 'threads' || collection === 'labels' || collection === 'drafts';
          };
        })(this));
        this.worker.resume();
        return expect(this.worker.fetchCollection.calls.map(function(call) {
          return call.args[0];
        })).toEqual(['threads', 'labels', 'drafts']);
      });
      return it("should be called when Actions.retrySync is received", function() {
        spyOn(this.worker, 'resume').andCallThrough();
        Actions.retrySync();
        return expect(this.worker.resume).toHaveBeenCalled();
      });
    });
    describe("shouldFetchCollection", function() {
      it("should return false if the collection sync is already in progress", function() {
        this.worker._state.threads = {
          'busy': true,
          'complete': false
        };
        return expect(this.worker.shouldFetchCollection('threads')).toBe(false);
      });
      it("should return false if the collection sync is already complete", function() {
        this.worker._state.threads = {
          'busy': false,
          'complete': true
        };
        return expect(this.worker.shouldFetchCollection('threads')).toBe(false);
      });
      return it("should return true otherwise", function() {
        this.worker._state.threads = {
          'busy': false,
          'complete': false
        };
        expect(this.worker.shouldFetchCollection('threads')).toBe(true);
        this.worker._state.threads = void 0;
        return expect(this.worker.shouldFetchCollection('threads')).toBe(true);
      });
    });
    describe("fetchCollection", function() {
      beforeEach(function() {
        return this.apiRequests = [];
      });
      it("should start the request for the model count", function() {
        this.worker._state.threads = {
          'busy': false,
          'complete': false
        };
        this.worker.fetchCollection('threads');
        expect(this.apiRequests[0].requestOptions.path).toBe('/threads');
        return expect(this.apiRequests[0].requestOptions.qs.view).toBe('count');
      });
      it("should pass any metadata it preloaded", function() {
        this.worker._state.threads = {
          'busy': false,
          'complete': false
        };
        this.worker.fetchCollection('threads');
        expect(this.apiRequests[1].model).toBe('threads');
        return expect(this.apiRequests[1].requestOptions.metadataToAttach).toBe(this.worker._metadata);
      });
      describe("when there is no request history (`lastRequestRange`)", function() {
        return it("should start the first request for models", function() {
          this.worker._state.threads = {
            'busy': false,
            'complete': false
          };
          this.worker.fetchCollection('threads');
          expect(this.apiRequests[1].model).toBe('threads');
          return expect(this.apiRequests[1].params.offset).toBe(0);
        });
      });
      describe("when it was previously trying to fetch a page (`lastRequestRange`)", function() {
        beforeEach(function() {
          return this.worker._state.threads = {
            'count': 1200,
            'fetched': 100,
            'busy': false,
            'complete': false,
            'error': new Error("Something bad"),
            'lastRequestRange': {
              offset: 100,
              limit: 50
            }
          };
        });
        it("should start paginating from the request that was interrupted", function() {
          this.worker.fetchCollection('threads');
          expect(this.apiRequests[0].model).toBe('threads');
          expect(this.apiRequests[0].params.offset).toBe(100);
          return expect(this.apiRequests[0].params.limit).toBe(50);
        });
        return it("should not reset the `count`, `fetched` or start fetching the count", function() {
          this.worker.fetchCollection('threads');
          expect(this.worker._state.threads.fetched).toBe(100);
          expect(this.worker._state.threads.count).toBe(1200);
          return expect(this.apiRequests.length).toBe(1);
        });
      });
      return describe('when maxFetchCount option is specified', function() {
        it("should only fetch maxFetch count on the first request if it is less than initialPageSize", function() {
          this.worker._state.messages = {
            count: 1000,
            fetched: 0
          };
          this.worker.fetchCollection('messages', {
            initialPageSize: 30,
            maxFetchCount: 25
          });
          expect(this.apiRequests[0].params.offset).toBe(0);
          return expect(this.apiRequests[0].params.limit).toBe(25);
        });
        return it("sould only fetch the maxFetchCount when restoring from saved state", function() {
          this.worker._state.messages = {
            count: 1000,
            fetched: 470,
            lastRequestRange: {
              limit: 50,
              offset: 470
            }
          };
          this.worker.fetchCollection('messages', {
            maxFetchCount: 500
          });
          expect(this.apiRequests[0].params.offset).toBe(470);
          return expect(this.apiRequests[0].params.limit).toBe(30);
        });
      });
    });
    describe("fetchCollectionPage", function() {
      beforeEach(function() {
        return this.apiRequests = [];
      });
      return describe('when maxFetchCount option is specified', function() {
        it('should not fetch next page if maxFetchCount has been reached', function() {
          var success;
          this.worker._state.messages = {
            count: 1000,
            fetched: 470
          };
          this.worker.fetchCollectionPage('messages', {
            limit: 30,
            offset: 470
          }, {
            maxFetchCount: 500
          });
          success = this.apiRequests[0].requestOptions.success;
          success({
            length: 30
          });
          expect(this.worker._state.messages.fetched).toBe(500);
          advanceClock(2000);
          return expect(this.apiRequests.length).toBe(1);
        });
        return it('should limit by maxFetchCount when requesting the next page', function() {
          var success;
          this.worker._state.messages = {
            count: 1000,
            fetched: 450
          };
          this.worker.fetchCollectionPage('messages', {
            limit: 30,
            offset: 450
          }, {
            maxFetchCount: 500
          });
          success = this.apiRequests[0].requestOptions.success;
          success({
            length: 30
          });
          expect(this.worker._state.messages.fetched).toBe(480);
          advanceClock(2000);
          expect(this.apiRequests[1].params.offset).toBe(480);
          return expect(this.apiRequests[1].params.limit).toBe(20);
        });
      });
    });
    describe("when an API request completes", function() {
      beforeEach(function() {
        this.worker.start();
        advanceClock();
        this.request = this.apiRequests[1];
        return this.apiRequests = [];
      });
      describe("successfully, with models", function() {
        it("should start out by requesting a small number of items", function() {
          return expect(this.request.params.limit).toBe(NylasSyncWorker.INITIAL_PAGE_SIZE);
        });
        it("should request the next page", function() {
          var i, models, pageSize, _i, _ref1;
          pageSize = this.request.params.limit;
          models = [];
          for (i = _i = 0, _ref1 = pageSize - 1; 0 <= _ref1 ? _i <= _ref1 : _i >= _ref1; i = 0 <= _ref1 ? ++_i : --_i) {
            models.push(new Thread);
          }
          this.request.requestOptions.success(models);
          advanceClock(2000);
          expect(this.apiRequests.length).toBe(1);
          return expect(this.apiRequests[0].params.offset).toEqual(this.request.params.offset + pageSize);
        });
        it("increase the limit on the next page load by 50%", function() {
          var i, models, pageSize, _i, _ref1;
          pageSize = this.request.params.limit;
          models = [];
          for (i = _i = 0, _ref1 = pageSize - 1; 0 <= _ref1 ? _i <= _ref1 : _i >= _ref1; i = 0 <= _ref1 ? ++_i : --_i) {
            models.push(new Thread);
          }
          this.request.requestOptions.success(models);
          advanceClock(2000);
          expect(this.apiRequests.length).toBe(1);
          return expect(this.apiRequests[0].params.limit).toEqual(pageSize * 1.5);
        });
        it("never requests more then MAX_PAGE_SIZE", function() {
          var i, models, pageSize, _i, _ref1;
          pageSize = this.request.params.limit = NylasSyncWorker.MAX_PAGE_SIZE;
          models = [];
          for (i = _i = 0, _ref1 = pageSize - 1; 0 <= _ref1 ? _i <= _ref1 : _i >= _ref1; i = 0 <= _ref1 ? ++_i : --_i) {
            models.push(new Thread);
          }
          this.request.requestOptions.success(models);
          advanceClock(2000);
          expect(this.apiRequests.length).toBe(1);
          return expect(this.apiRequests[0].params.limit).toEqual(NylasSyncWorker.MAX_PAGE_SIZE);
        });
        return it("should update the fetched count on the collection", function() {
          var i, models, pageSize, _i, _ref1;
          expect(this.worker.state().threads.fetched).toEqual(0);
          pageSize = this.request.params.limit;
          models = [];
          for (i = _i = 0, _ref1 = pageSize - 1; 0 <= _ref1 ? _i <= _ref1 : _i >= _ref1; i = 0 <= _ref1 ? ++_i : --_i) {
            models.push(new Thread);
          }
          this.request.requestOptions.success(models);
          return expect(this.worker.state().threads.fetched).toEqual(pageSize);
        });
      });
      describe("successfully, with fewer models than requested", function() {
        beforeEach(function() {
          var i, models, _i;
          models = [];
          for (i = _i = 0; _i <= 100; i = ++_i) {
            models.push(new Thread);
          }
          return this.request.requestOptions.success(models);
        });
        it("should not request another page", function() {
          return expect(this.apiRequests.length).toBe(0);
        });
        it("should update the state to complete", function() {
          expect(this.worker.state().threads.busy).toEqual(false);
          return expect(this.worker.state().threads.complete).toEqual(true);
        });
        return it("should update the fetched count on the collection", function() {
          return expect(this.worker.state().threads.fetched).toEqual(101);
        });
      });
      describe("successfully, with no models", function() {
        it("should not request another page", function() {
          this.request.requestOptions.success([]);
          return expect(this.apiRequests.length).toBe(0);
        });
        return it("should update the state to complete", function() {
          this.request.requestOptions.success([]);
          expect(this.worker.state().threads.busy).toEqual(false);
          return expect(this.worker.state().threads.complete).toEqual(true);
        });
      });
      describe("with an error", function() {
        it("should log the error to the state, along with the range that failed", function() {
          var err;
          err = new Error("Oh no a network error");
          this.request.requestOptions.error(err);
          expect(this.worker.state().threads.busy).toEqual(false);
          expect(this.worker.state().threads.complete).toEqual(false);
          expect(this.worker.state().threads.error).toEqual(err.toString());
          return expect(this.worker.state().threads.lastRequestRange).toEqual({
            offset: 0,
            limit: 30
          });
        });
        return it("should not request another page", function() {
          this.request.requestOptions.error(new Error("Oh no a network error"));
          return expect(this.apiRequests.length).toBe(0);
        });
      });
      return describe("succeeds after a previous error", function() {
        beforeEach(function() {
          this.worker._state.threads.error = new Error("Something bad happened");
          this.worker._state.threads.lastRequestRange = {
            limit: 10,
            offset: 10
          };
          this.request.requestOptions.success([]);
          return advanceClock(1);
        });
        return it("should clear any previous error and updates lastRequestRange", function() {
          expect(this.worker.state().threads.error).toEqual(null);
          return expect(this.worker.state().threads.lastRequestRange).toEqual({
            offset: 0,
            limit: 30
          });
        });
      });
    });
    return describe("cleanup", function() {
      it("should termiate the delta connection", function() {
        spyOn(this.connection, 'end');
        this.worker.cleanup();
        return expect(this.connection.end).toHaveBeenCalled();
      });
      return it("should stop trying to restart failed collection syncs", function() {
        spyOn(console, 'log');
        spyOn(this.worker, 'resume').andCallThrough();
        this.worker.cleanup();
        advanceClock(50000);
        return expect(this.worker.resume.callCount).toBe(0);
      });
    });
  });

}).call(this);
