(function() {
  var APMWrapper, BufferedProcess, Q, createJsonParseError, createProcessError, handleProcessErrors, semver, _;

  _ = require('underscore');

  Q = require('q');

  semver = require('semver');

  BufferedProcess = require('./buffered-process');

  module.exports = APMWrapper = (function() {
    function APMWrapper() {
      this.packagePromises = [];
    }

    APMWrapper.prototype.runCommand = function(args, options, callback) {
      var command, errorLines, exit, outputLines, stderr, stdout;
      command = NylasEnv.packages.getApmPath();
      outputLines = [];
      stdout = function(lines) {
        return outputLines.push(lines);
      };
      errorLines = [];
      stderr = function(lines) {
        return errorLines.push(lines);
      };
      exit = function(code) {
        return callback(code, outputLines.join('\n'), errorLines.join('\n'));
      };
      options || (options = {});
      options.env = {
        ATOM_API_URL: 'https://edgehill-packages.nylas.com/api',
        ATOM_RESOURCE_PATH: NylasEnv.getLoadSettings().resourcePath,
        ATOM_HOME: NylasEnv.getConfigDirPath()
      };
      if (process.platform === "win32") {
        options.env["ProgramFiles"] = process.env.ProgramFiles;
      }
      args.push('--no-color');
      return new BufferedProcess({
        command: command,
        args: args,
        stdout: stdout,
        stderr: stderr,
        exit: exit,
        options: options
      });
    };

    APMWrapper.prototype.runCommandReturningPackages = function(args, errorMessage, callback) {
      return this.runCommand(args, null, function(code, stdout, stderr) {
        var error, packages, parseError, _ref;
        if (code === 0) {
          try {
            packages = (_ref = JSON.parse(stdout)) != null ? _ref : [];
          } catch (_error) {
            parseError = _error;
            error = createJsonParseError(errorMessage, parseError, stdout);
            return callback(error);
          }
          return callback(null, packages);
        } else {
          error = new Error(errorMessage);
          error.stdout = stdout;
          error.stderr = stderr;
          return callback(error);
        }
      });
    };

    APMWrapper.prototype.loadInstalled = function(callback) {
      var apmProcess, args, errorMessage;
      args = ['ls', '--json'];
      errorMessage = 'Fetching local packages failed.';
      apmProcess = this.runCommandReturningPackages(args, errorMessage, callback);
      return handleProcessErrors(apmProcess, errorMessage, callback);
    };

    APMWrapper.prototype.loadFeatured = function(options, callback) {
      var apmProcess, args, errorMessage, version;
      args = ['featured', '--json'];
      version = NylasEnv.getVersion();
      if (options.themes) {
        args.push('--themes');
      }
      if (semver.valid(version)) {
        args.push('--compatible', version);
      }
      errorMessage = 'Fetching featured packages failed.';
      apmProcess = this.runCommandReturningPackages(args, errorMessage, callback);
      return handleProcessErrors(apmProcess, errorMessage, callback);
    };

    APMWrapper.prototype.loadOutdated = function(callback) {
      var apmProcess, args, errorMessage, version;
      args = ['outdated', '--json'];
      version = NylasEnv.getVersion();
      if (semver.valid(version)) {
        args.push('--compatible', version);
      }
      errorMessage = 'Fetching outdated packages and themes failed.';
      apmProcess = this.runCommandReturningPackages(args, errorMessage, callback);
      return handleProcessErrors(apmProcess, errorMessage, callback);
    };

    APMWrapper.prototype.loadPackage = function(packageName, callback) {
      var apmProcess, args, errorMessage;
      args = ['view', packageName, '--json'];
      errorMessage = "Fetching package '" + packageName + "' failed.";
      apmProcess = this.runCommandReturningPackages(args, errorMessage, callback);
      return handleProcessErrors(apmProcess, errorMessage, callback);
    };

    APMWrapper.prototype.loadCompatiblePackageVersion = function(packageName, callback) {
      var apmProcess, args, errorMessage;
      args = ['view', packageName, '--json', '--compatible', this.normalizeVersion(NylasEnv.getVersion())];
      errorMessage = "Fetching package '" + packageName + "' failed.";
      apmProcess = this.runCommandReturningPackages(args, errorMessage, callback);
      return handleProcessErrors(apmProcess, errorMessage, callback);
    };

    APMWrapper.prototype.getInstalled = function() {
      return Promise.promisify(this.loadInstalled, this)();
    };

    APMWrapper.prototype.getFeatured = function(options) {
      if (options == null) {
        options = {};
      }
      return Promise.promisify(this.loadFeatured, this)(options);
    };

    APMWrapper.prototype.getOutdated = function() {
      return Promise.promisify(this.loadOutdated, this)();
    };

    APMWrapper.prototype.getPackage = function(packageName) {
      var _base;
      return (_base = this.packagePromises)[packageName] != null ? _base[packageName] : _base[packageName] = Promise.promisify(this.loadPackage, this, packageName)();
    };

    APMWrapper.prototype.satisfiesVersion = function(version, metadata) {
      var engine, _ref, _ref1;
      engine = (_ref = (_ref1 = metadata.engines) != null ? _ref1.nylas : void 0) != null ? _ref : '*';
      if (!semver.validRange(engine)) {
        return false;
      }
      return semver.satisfies(version, engine);
    };

    APMWrapper.prototype.normalizeVersion = function(version) {
      if (typeof version === 'string') {
        version = version.split('-')[0];
      }
      return version;
    };

    APMWrapper.prototype.search = function(query, options) {
      var apmProcess, args, deferred, errorMessage;
      if (options == null) {
        options = {};
      }
      deferred = Promise.defer();
      args = ['search', query, '--json'];
      if (options.themes) {
        args.push('--themes');
      } else if (options.packages) {
        args.push('--packages');
      }
      errorMessage = "Searching for \u201C" + query + "\u201D failed.";
      apmProcess = this.runCommand(args, null, function(code, stdout, stderr) {
        var error, packages, parseError, _ref;
        if (code === 0) {
          try {
            packages = (_ref = JSON.parse(stdout)) != null ? _ref : [];
            return deferred.resolve(packages);
          } catch (_error) {
            parseError = _error;
            error = createJsonParseError(errorMessage, parseError, stdout);
            return deferred.reject(error);
          }
        } else {
          error = new Error(errorMessage);
          error.stdout = stdout;
          error.stderr = stderr;
          return deferred.reject(error);
        }
      });
      handleProcessErrors(apmProcess, errorMessage, function(error) {
        return deferred.reject(error);
      });
      return deferred.promise;
    };

    APMWrapper.prototype.update = function(pack, newVersion, callback) {
      var activateOnFailure, activateOnSuccess, apmProcess, args, errorMessage, exit, name, onError, theme;
      name = pack.name, theme = pack.theme;
      if (theme) {
        activateOnSuccess = NylasEnv.packages.isPackageActive(name);
      } else {
        activateOnSuccess = !NylasEnv.packages.isPackageDisabled(name);
      }
      activateOnFailure = NylasEnv.packages.isPackageActive(name);
      if (NylasEnv.packages.isPackageActive(name)) {
        NylasEnv.packages.deactivatePackage(name);
      }
      if (NylasEnv.packages.isPackageLoaded(name)) {
        NylasEnv.packages.unloadPackage(name);
      }
      errorMessage = "Updating to \u201C" + name + "@" + newVersion + "\u201D failed.";
      onError = (function(_this) {
        return function(error) {
          error.packageInstallError = !theme;
          return callback(error);
        };
      })(this);
      args = ['install', "" + name + "@" + newVersion];
      exit = (function(_this) {
        return function(code, stdout, stderr) {
          var error;
          if (code === 0) {
            if (activateOnSuccess) {
              NylasEnv.packages.activatePackage(name);
            } else {
              NylasEnv.packages.loadPackage(name);
            }
            return typeof callback === "function" ? callback() : void 0;
          } else {
            if (activateOnFailure) {
              NylasEnv.packages.activatePackage(name);
            }
            error = new Error(errorMessage);
            error.stdout = stdout;
            error.stderr = stderr;
            return onError(error);
          }
        };
      })(this);
      apmProcess = this.runCommand(args, null, exit);
      return handleProcessErrors(apmProcess, errorMessage, onError);
    };

    APMWrapper.prototype.unload = function(packageName) {
      if (NylasEnv.packages.isPackageLoaded(name)) {
        if (NylasEnv.packages.isPackageActive(name)) {
          NylasEnv.packages.deactivatePackage(name);
        }
        return NylasEnv.packages.unloadPackage(name);
      }
    };

    APMWrapper.prototype.install = function(pack, callback) {
      var activateOnFailure, activateOnSuccess, apmProcess, args, errorMessage, exit, name, onError, theme, version;
      name = pack.name, version = pack.version, theme = pack.theme;
      activateOnSuccess = !theme && !NylasEnv.packages.isPackageDisabled(name);
      activateOnFailure = NylasEnv.packages.isPackageActive(name);
      this.unload(name);
      args = ['install', "" + name + "@" + version];
      errorMessage = "Installing \u201C" + name + "@" + version + "\u201D failed.";
      onError = (function(_this) {
        return function(error) {
          error.packageInstallError = !theme;
          return callback(error);
        };
      })(this);
      exit = (function(_this) {
        return function(code, stdout, stderr) {
          var error;
          if (code === 0) {
            if (activateOnSuccess) {
              NylasEnv.packages.activatePackage(name);
            } else {
              NylasEnv.packages.loadPackage(name);
            }
            return typeof callback === "function" ? callback() : void 0;
          } else {
            if (activateOnFailure) {
              NylasEnv.packages.activatePackage(name);
            }
            error = new Error(errorMessage);
            error.stdout = stdout;
            error.stderr = stderr;
            return onError(error);
          }
        };
      })(this);
      apmProcess = this.runCommand(args, null, exit);
      return handleProcessErrors(apmProcess, errorMessage, onError);
    };

    APMWrapper.prototype.installDependenciesInPackageDirectory = function(dir, callback) {
      var apmProcess, errorMessage, exit;
      errorMessage = "Running apm install failed to install package dependencies.";
      exit = (function(_this) {
        return function(code, stdout, stderr) {
          var error;
          if (code === 0) {
            return typeof callback === "function" ? callback() : void 0;
          } else {
            error = new Error(errorMessage);
            error.stdout = stdout;
            error.stderr = stderr;
            return callback(error);
          }
        };
      })(this);
      apmProcess = this.runCommand(['install'], {
        cwd: dir
      }, exit);
      return handleProcessErrors(apmProcess, errorMessage, callback);
    };

    APMWrapper.prototype.uninstall = function(pack, callback) {
      var apmProcess, errorMessage, name, onError;
      name = pack.name;
      if (NylasEnv.packages.isPackageActive(name)) {
        NylasEnv.packages.deactivatePackage(name);
      }
      errorMessage = "Uninstalling \u201C" + name + "\u201D failed.";
      onError = (function(_this) {
        return function(error) {
          return callback(error);
        };
      })(this);
      apmProcess = this.runCommand(['uninstall', '--hard', name], null, (function(_this) {
        return function(code, stdout, stderr) {
          var error;
          if (code === 0) {
            _this.unload(name);
            return typeof callback === "function" ? callback() : void 0;
          } else {
            error = new Error(errorMessage);
            error.stdout = stdout;
            error.stderr = stderr;
            return onError(error);
          }
        };
      })(this));
      return handleProcessErrors(apmProcess, errorMessage, onError);
    };

    APMWrapper.prototype.canUpgrade = function(installedPackage, availableVersion) {
      var installedVersion;
      if (installedPackage == null) {
        return false;
      }
      installedVersion = installedPackage.metadata.version;
      if (!semver.valid(installedVersion)) {
        return false;
      }
      if (!semver.valid(availableVersion)) {
        return false;
      }
      return semver.gt(availableVersion, installedVersion);
    };

    APMWrapper.prototype.checkNativeBuildTools = function() {
      var apmProcess, deferred;
      deferred = Promise.defer();
      apmProcess = this.runCommand(['install', '--check'], null, function(code, stdout, stderr) {
        if (code === 0) {
          return deferred.resolve();
        } else {
          return deferred.reject(new Error());
        }
      });
      apmProcess.onWillThrowError(function(_arg) {
        var error, handle;
        error = _arg.error, handle = _arg.handle;
        handle();
        return deferred.reject(error);
      });
      return deferred.promise;
    };

    return APMWrapper;

  })();

  createJsonParseError = function(message, parseError, stdout) {
    var error;
    error = new Error(message);
    error.stdout = '';
    error.stderr = "" + parseError.message + ": " + stdout;
    return error;
  };

  createProcessError = function(message, processError) {
    var error;
    error = new Error(message);
    error.stdout = '';
    error.stderr = processError.message;
    return error;
  };

  handleProcessErrors = function(apmProcess, message, callback) {
    return apmProcess.onWillThrowError(function(_arg) {
      var error, handle;
      error = _arg.error, handle = _arg.handle;
      handle();
      return callback(createProcessError(message, error));
    });
  };

}).call(this);
