(function() {
  var CoffeeHelpers, ComponentRegistry, DeprecatedRoles, Listener, Publisher, _, _ref,
    __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; };

  _ = require('underscore');

  _ref = require('./flux/modules/reflux-coffee'), Listener = _ref.Listener, Publisher = _ref.Publisher;

  CoffeeHelpers = require('./flux/coffee-helpers');

  DeprecatedRoles = {
    'thread:BulkAction': 'ThreadActionsToolbarButton',
    'draft:BulkAction': 'DraftActionsToolbarButton',
    'message:Toolbar': 'ThreadActionsToolbarButton',
    'thread:Toolbar': 'ThreadActionsToolbarButton'
  };


  /*
  Public: The ComponentRegistry maintains an index of React components registered
  by Nylas packages. Components can use {InjectedComponent} and {InjectedComponentSet}
  to dynamically render components registered with the ComponentRegistry.
  
  Section: Stores
   */

  ComponentRegistry = (function() {
    ComponentRegistry.include = CoffeeHelpers.includeModule;

    ComponentRegistry.include(Publisher);

    ComponentRegistry.include(Listener);

    function ComponentRegistry() {
      this.showComponentRegions = __bind(this.showComponentRegions, this);
      this._clear = __bind(this._clear, this);
      this.findComponentsMatching = __bind(this.findComponentsMatching, this);
      this.findComponentByName = __bind(this.findComponentByName, this);
      this.unregister = __bind(this.unregister, this);
      this.register = __bind(this.register, this);
      this._registry = {};
      this._cache = {};
      this._showComponentRegions = false;
    }

    ComponentRegistry.prototype.register = function(component, options) {
      var locations, modes, roles, _ref1;
      if (component.view != null) {
        return console.warn("Ignoring component trying to register with old CommandRegistry.register syntax");
      }
      if (!options) {
        throw new Error("ComponentRegistry.register() requires `options` that describe the component");
      }
      if (!component) {
        throw new Error("ComponentRegistry.register() requires `component`, a React component");
      }
      if (!component.displayName) {
        throw new Error("ComponentRegistry.register() requires that your React Component defines a `displayName`");
      }
      _ref1 = this._pluralizeDescriptor(options), locations = _ref1.locations, modes = _ref1.modes, roles = _ref1.roles;
      if (!roles && !locations) {
        throw new Error("ComponentRegistry.register() requires `role` or `location`");
      }
      if (this._registry[component.displayName] && this._registry[component.displayName].component !== component) {
        throw new Error("ComponentRegistry.register(): A different component was already registered with the name " + component.displayName);
      }
      if (roles) {
        roles = this._removeDeprecatedRoles(component.displayName, roles);
      }
      this._cache = {};
      this._registry[component.displayName] = {
        component: component,
        locations: locations,
        modes: modes,
        roles: roles
      };
      this.triggerDebounced();
      return this;
    };

    ComponentRegistry.prototype.unregister = function(component) {
      if (_.isString(component)) {
        throw new Error("ComponentRegistry.unregister() must be called with a component.");
      }
      this._cache = {};
      delete this._registry[component.displayName];
      return this.triggerDebounced();
    };

    ComponentRegistry.prototype.findComponentByName = function(name) {
      var _ref1;
      return (_ref1 = this._registry[name]) != null ? _ref1.component : void 0;
    };


    /*
    Public: Retrieve all of the registry entries matching a given descriptor.
    
    ```coffee
      ComponentRegistry.findComponentsMatching({
        role: 'Composer:ActionButton'
      })
    
      ComponentRegistry.findComponentsMatching({
        location: WorkspaceStore.Location.RootSidebar.Toolbar
      })
    ```
    
    - `descriptor`: An {Object} that specifies set of components using the
      available keys below.
    
      * `mode`: (optional) {String} Components that specifically list modes
         will only be returned if they include this mode.
    
      * `role`: (optional) {String} Only return components that have registered
         for this role.
    
      * `location`: (optional) {Object} Only return components that have registered
         for this location.
    
      Note that for advanced use cases, you can also pass (`modes`, `roles`, `locations`)
      with arrays instead of single values.
    
    Returns an {Array} of {React.Component} objects
     */

    ComponentRegistry.prototype.findComponentsMatching = function(descriptor) {
      var cacheKey, entries, locations, modes, overlaps, results, roles, _ref1;
      if (descriptor == null) {
        throw new Error("ComponentRegistry.findComponentsMatching called without descriptor");
      }
      _ref1 = this._pluralizeDescriptor(descriptor), locations = _ref1.locations, modes = _ref1.modes, roles = _ref1.roles;
      if (!locations && !modes && !roles) {
        throw new Error("ComponentRegistry.findComponentsMatching called with an empty descriptor");
      }
      cacheKey = JSON.stringify({
        locations: locations,
        modes: modes,
        roles: roles
      });
      if (this._cache[cacheKey]) {
        return [].concat(this._cache[cacheKey]);
      }
      overlaps = function(entry, search) {
        if (entry == null) {
          entry = [];
        }
        if (search == null) {
          search = [];
        }
        return _.intersection(entry, search).length > 0;
      };
      entries = _.values(this._registry);
      entries = _.filter(entries, function(entry) {
        if (modes && entry.modes && !overlaps(modes, entry.modes)) {
          return false;
        }
        if (locations && !overlaps(locations, entry.locations)) {
          return false;
        }
        if (roles && !overlaps(roles, entry.roles)) {
          return false;
        }
        return true;
      });
      results = _.map(entries, function(entry) {
        return entry.component;
      });
      this._cache[cacheKey] = results;
      return [].concat(results);
    };

    ComponentRegistry.prototype.triggerDebounced = _.debounce((function() {
      return this.trigger(this);
    }), 16);

    ComponentRegistry.prototype._removeDeprecatedRoles = function(displayName, roles) {
      var newRoles;
      newRoles = _.clone(roles);
      roles.forEach(function(role, idx) {
        var instead;
        if (role in DeprecatedRoles) {
          instead = DeprecatedRoles[role];
          console.warn("Deprecation warning! The role `" + role + "` has been deprecated. Register `" + displayName + "` for the role `" + instead + "` instead.");
          return newRoles.splice(idx, 1, instead);
        }
      });
      return newRoles;
    };

    ComponentRegistry.prototype._pluralizeDescriptor = function(descriptor) {
      var locations, modes, roles;
      locations = descriptor.locations, modes = descriptor.modes, roles = descriptor.roles;
      if (descriptor.mode) {
        modes = [descriptor.mode];
      }
      if (descriptor.role) {
        roles = [descriptor.role];
      }
      if (descriptor.location) {
        locations = [descriptor.location];
      }
      return {
        locations: locations,
        modes: modes,
        roles: roles
      };
    };

    ComponentRegistry.prototype._clear = function() {
      this._cache = {};
      return this._registry = {};
    };

    ComponentRegistry.prototype.toggleComponentRegions = function() {
      this._showComponentRegions = !this._showComponentRegions;
      return this.trigger(this);
    };

    ComponentRegistry.prototype.showComponentRegions = function() {
      return this._showComponentRegions;
    };

    return ComponentRegistry;

  })();

  module.exports = new ComponentRegistry();

}).call(this);
