(function() {
  var Actions, ComponentRegistry, Flexbox, InjectedComponentLabel, InjectedComponentSet, React, UnsafeComponent, Utils, WorkspaceStore, _, _ref,
    __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; },
    __hasProp = {}.hasOwnProperty,
    __extends = function(child, parent) { for (var key in parent) { if (__hasProp.call(parent, key)) child[key] = parent[key]; } function ctor() { this.constructor = child; } ctor.prototype = parent.prototype; child.prototype = new ctor(); child.__super__ = parent.prototype; return child; };

  React = require('react');

  _ = require('underscore');

  UnsafeComponent = require('./unsafe-component');

  Flexbox = require('./flexbox');

  InjectedComponentLabel = require('./injected-component-label');

  _ref = require("nylas-exports"), Utils = _ref.Utils, Actions = _ref.Actions, WorkspaceStore = _ref.WorkspaceStore, ComponentRegistry = _ref.ComponentRegistry;


  /*
  Public: InjectedComponent makes it easy to include a set of dynamically registered
  components inside of your React render method. Rather than explicitly render
  an array of buttons, for example, you can use InjectedComponentSet:
  
  ```coffee
  <InjectedComponentSet className="message-actions"
                    matching={role: 'ThreadActionButton'}
                    exposedProps={thread:@props.thread, message:@props.message}>
  ```
  
  InjectedComponentSet will look up components registered for the location you provide,
  render them inside a {Flexbox} and pass them `exposedProps`. By default, all injected
  children are rendered inside {UnsafeComponent} wrappers to prevent third-party code
  from throwing exceptions that break React renders.
  
  InjectedComponentSet monitors the ComponentRegistry for changes. If a new component
  is registered into the location you provide, InjectedComponentSet will re-render.
  
  If no matching components is found, the InjectedComponent renders an empty span.
  
  Section: Component Kit
   */

  InjectedComponentSet = (function(_super) {
    __extends(InjectedComponentSet, _super);

    InjectedComponentSet.displayName = 'InjectedComponentSet';


    /*
    Public: React `props` supported by InjectedComponentSet:
    
     - `matching` Pass an {Object} with ComponentRegistry descriptors
        This set of descriptors is provided to {ComponentRegistry::findComponentsForDescriptor}
        to retrieve components for display.
     - `matchLimit` (optional) A {Number} that indicates the max number of matching elements to render
     - `className` (optional) A {String} class name for the containing element.
     - `children` (optional) Any React elements rendered inside the InjectedComponentSet
        will always be displayed.
     - `onComponentsDidRender` Callback that will be called when the injected component set
        is successfully rendered onto the DOM.
     - `exposedProps` (optional) An {Object} with props that will be passed to each
        item rendered into the set.
     - `containersRequired` (optional). Pass false to optionally remove the containers
        placed around injected components to isolate them from the rest of the app.
    
     -  Any other props you provide, such as `direction`, `data-column`, etc.
        will be applied to the {Flexbox} rendered by the InjectedComponentSet.
     */

    InjectedComponentSet.propTypes = {
      matching: React.PropTypes.object.isRequired,
      children: React.PropTypes.array,
      className: React.PropTypes.string,
      matchLimit: React.PropTypes.number,
      exposedProps: React.PropTypes.object,
      containersRequired: React.PropTypes.bool,
      onComponentsDidRender: React.PropTypes.func
    };

    InjectedComponentSet.defaultProps = {
      direction: 'row',
      containersRequired: true,
      onComponentsDidRender: function() {}
    };

    function InjectedComponentSet(props) {
      this.props = props;
      this._getStateFromStores = __bind(this._getStateFromStores, this);
      this._onComponentDidRender = __bind(this._onComponentDidRender, this);
      this.render = __bind(this.render, this);
      this.componentDidUpdate = __bind(this.componentDidUpdate, this);
      this.componentWillReceiveProps = __bind(this.componentWillReceiveProps, this);
      this.componentWillUnmount = __bind(this.componentWillUnmount, this);
      this.componentDidMount = __bind(this.componentDidMount, this);
      this.state = this._getStateFromStores();
      this._renderedComponents = new Set();
    }

    InjectedComponentSet.prototype.componentDidMount = function() {
      this._componentUnlistener = ComponentRegistry.listen((function(_this) {
        return function() {
          return _this.setState(_this._getStateFromStores());
        };
      })(this));
      if (this.props.containersRequired === false) {
        return this.props.onComponentsDidRender();
      }
    };

    InjectedComponentSet.prototype.componentWillUnmount = function() {
      if (this._componentUnlistener) {
        return this._componentUnlistener();
      }
    };

    InjectedComponentSet.prototype.componentWillReceiveProps = function(newProps) {
      var _ref1;
      if (newProps.location !== ((_ref1 = this.props) != null ? _ref1.location : void 0)) {
        return this.setState(this._getStateFromStores(newProps));
      }
    };

    InjectedComponentSet.prototype.componentDidUpdate = function() {
      if (this.props.containersRequired === false) {
        return this.props.onComponentsDidRender();
      }
    };

    InjectedComponentSet.prototype.render = function() {
      var elements, exposedProps, flexboxClassName, flexboxProps, _ref1, _ref2, _ref3;
      this._renderedComponents = new Set();
      flexboxProps = Utils.fastOmit(this.props, Object.keys(this.constructor.propTypes));
      flexboxClassName = (_ref1 = this.props.className) != null ? _ref1 : "";
      exposedProps = (_ref2 = this.props.exposedProps) != null ? _ref2 : {};
      elements = this.state.components.map((function(_this) {
        return function(component) {
          if (_this.props.containersRequired === false || component.containerRequired === false) {
            return React.createElement(component, React.__spread({
              "key": component.displayName
            }, exposedProps));
          } else {
            return React.createElement(UnsafeComponent, React.__spread({
              "key": component.displayName,
              "component": component,
              "onComponentDidRender": _this._onComponentDidRender.bind(_this, component.displayName)
            }, exposedProps));
          }
        };
      })(this));
      if (this.state.visible) {
        flexboxClassName += " registered-region-visible";
        elements.splice(0, 0, React.createElement(InjectedComponentLabel, React.__spread({
          "key": "_label",
          "matching": this.props.matching
        }, exposedProps)));
        elements.push(React.createElement("span", {
          "key": "_clear",
          "style": {
            clear: 'both'
          }
        }));
      }
      return React.createElement(Flexbox, React.__spread({
        "className": flexboxClassName
      }, flexboxProps), elements, (_ref3 = this.props.children) != null ? _ref3 : []);
    };

    InjectedComponentSet.prototype._onComponentDidRender = function(componentName) {
      this._renderedComponents.add(componentName);
      if (this._renderedComponents.size === this.state.components.length) {
        return this.props.onComponentsDidRender();
      }
    };

    InjectedComponentSet.prototype._getStateFromStores = function(props) {
      var limit;
      if (props == null) {
        props = this.props;
      }
      limit = props.matchLimit;
      return {
        components: ComponentRegistry.findComponentsMatching(this.props.matching).slice(0, limit),
        visible: ComponentRegistry.showComponentRegions()
      };
    };

    return InjectedComponentSet;

  })(React.Component);

  module.exports = InjectedComponentSet;

}).call(this);
