(function() {
  var Actions, ComponentRegistry, InjectedComponent, InjectedComponentLabel, React, ReactDOM, UnsafeComponent, WorkspaceStore, _, _ref,
    __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; },
    __hasProp = {}.hasOwnProperty,
    __extends = function(child, parent) { for (var key in parent) { if (__hasProp.call(parent, key)) child[key] = parent[key]; } function ctor() { this.constructor = child; } ctor.prototype = parent.prototype; child.prototype = new ctor(); child.__super__ = parent.prototype; return child; },
    __slice = [].slice;

  React = require('react');

  ReactDOM = require('react-dom');

  _ = require('underscore');

  UnsafeComponent = require('./unsafe-component');

  InjectedComponentLabel = require('./injected-component-label');

  _ref = require("nylas-exports"), Actions = _ref.Actions, WorkspaceStore = _ref.WorkspaceStore, ComponentRegistry = _ref.ComponentRegistry;


  /*
  Public: InjectedComponent makes it easy to include dynamically registered
  components inside of your React render method. Rather than explicitly render
  a component, such as a `<Composer>`, you can use InjectedComponent:
  
  ```coffee
  <InjectedComponent matching={role:"Composer"} exposedProps={draftClientId:123} />
  ```
  
  InjectedComponent will look up the component registered with that role in the
  {ComponentRegistry} and render it, passing the exposedProps (`draftClientId={123}`) along.
  
  InjectedComponent monitors the ComponentRegistry for changes. If a new component
  is registered that matches the descriptor you provide, InjectedComponent will refresh.
  
  If no matching component is found, the InjectedComponent renders an empty div.
  
  Section: Component Kit
   */

  InjectedComponent = (function(_super) {
    __extends(InjectedComponent, _super);

    InjectedComponent.displayName = 'InjectedComponent';


    /*
    Public: React `props` supported by InjectedComponent:
    
     - `matching` Pass an {Object} with ComponentRegistry descriptors.
        This set of descriptors is provided to {ComponentRegistry::findComponentsForDescriptor}
        to retrieve the component that will be displayed.
    
     - `onComponentDidRender` (optional) Callback that will be called when the injected component
        is successfully rendered onto the DOM.
    
     - `className` (optional) A {String} class name for the containing element.
    
     - `exposedProps` (optional) An {Object} with props that will be passed to each
        item rendered into the set.
    
     - `fallback` (optional) A {Component} to default to in case there are no matching
       components in the ComponentRegistry
    
     - `requiredMethods` (options) An {Array} with a list of methods that should be
       implemented by the registered component instance. If these are not implemented,
       an error will be thrown.
     */

    InjectedComponent.propTypes = {
      matching: React.PropTypes.object.isRequired,
      className: React.PropTypes.string,
      exposedProps: React.PropTypes.object,
      fallback: React.PropTypes.func,
      onComponentDidRender: React.PropTypes.func,
      style: React.PropTypes.object,
      requiredMethods: React.PropTypes.arrayOf(React.PropTypes.string)
    };

    InjectedComponent.defaultProps = {
      style: {},
      exposedProps: {},
      requiredMethods: [],
      onComponentDidRender: function() {}
    };

    function InjectedComponent(props) {
      this.props = props;
      this._getStateFromStores = __bind(this._getStateFromStores, this);
      this._verifyRequiredMethods = __bind(this._verifyRequiredMethods, this);
      this._setRequiredMethods = __bind(this._setRequiredMethods, this);
      this._runInnerDOMMethod = __bind(this._runInnerDOMMethod, this);
      this.blur = __bind(this.blur, this);
      this.focus = __bind(this.focus, this);
      this.render = __bind(this.render, this);
      this.componentDidUpdate = __bind(this.componentDidUpdate, this);
      this.componentWillReceiveProps = __bind(this.componentWillReceiveProps, this);
      this.componentWillUnmount = __bind(this.componentWillUnmount, this);
      this.componentDidMount = __bind(this.componentDidMount, this);
      this.state = this._getStateFromStores();
      this._verifyRequiredMethods();
      this._setRequiredMethods(this.props.requiredMethods);
    }

    InjectedComponent.prototype.componentDidMount = function() {
      var _ref1;
      this._componentUnlistener = ComponentRegistry.listen((function(_this) {
        return function() {
          return _this.setState(_this._getStateFromStores());
        };
      })(this));
      if (((_ref1 = this.state.component) != null ? _ref1.containerRequired : void 0) === false) {
        return this.props.onComponentDidRender();
      }
    };

    InjectedComponent.prototype.componentWillUnmount = function() {
      if (this._componentUnlistener) {
        return this._componentUnlistener();
      }
    };

    InjectedComponent.prototype.componentWillReceiveProps = function(newProps) {
      var _ref1;
      if (!_.isEqual(newProps.matching, (_ref1 = this.props) != null ? _ref1.matching : void 0)) {
        return this.setState(this._getStateFromStores(newProps));
      }
    };

    InjectedComponent.prototype.componentDidUpdate = function() {
      var _ref1;
      this._setRequiredMethods(this.props.requiredMethods);
      if (((_ref1 = this.state.component) != null ? _ref1.containerRequired : void 0) === false) {
        return this.props.onComponentDidRender();
      }
    };

    InjectedComponent.prototype.render = function() {
      var className, component, element, exposedProps, _ref1;
      if (!this.state.component) {
        return React.createElement("div", null);
      }
      exposedProps = Object.assign({}, this.props.exposedProps, {
        fallback: this.props.fallback
      });
      className = (_ref1 = this.props.className) != null ? _ref1 : "";
      if (this.state.visible) {
        className += " registered-region-visible";
      }
      component = this.state.component;
      if (component.containerRequired === false) {
        element = React.createElement(component, React.__spread({
          "ref": "inner",
          "key": component.displayName
        }, exposedProps));
      } else {
        element = React.createElement(UnsafeComponent, React.__spread({
          "ref": "inner",
          "style": this.props.style,
          "className": className,
          "key": component.displayName,
          "component": component,
          "onComponentDidRender": this.props.onComponentDidRender
        }, exposedProps));
      }
      if (this.state.visible) {
        return React.createElement("div", {
          "className": className,
          "style": this.props.style
        }, element, React.createElement(InjectedComponentLabel, React.__spread({
          "matching": this.props.matching
        }, exposedProps)), React.createElement("span", {
          "style": {
            clear: 'both'
          }
        }));
      } else {
        return React.createElement("div", {
          "className": className,
          "style": this.props.style
        }, element);
      }
    };

    InjectedComponent.prototype.focus = function() {
      return this._runInnerDOMMethod('focus');
    };

    InjectedComponent.prototype.blur = function() {
      return this._runInnerDOMMethod('blur');
    };

    InjectedComponent.prototype._runInnerDOMMethod = function() {
      var args, method, target;
      method = arguments[0], args = 2 <= arguments.length ? __slice.call(arguments, 1) : [];
      target = null;
      if (this.refs.inner instanceof UnsafeComponent && (this.refs.inner.injected[method] != null)) {
        target = this.refs.inner.injected;
      } else if (this.refs.inner && (this.refs.inner[method] != null)) {
        target = this.refs.inner;
      } else if (this.refs.inner) {
        target = ReactDOM.findDOMNode(this.refs.inner);
      } else {
        target = ReactDOM.findDOMNode(this);
      }
      return target[method].bind(target).apply(null, args);
    };

    InjectedComponent.prototype._setRequiredMethods = function(methods) {
      return methods.forEach((function(_this) {
        return function(method) {
          return Object.defineProperty(_this, method, {
            configurable: true,
            enumerable: true,
            value: function() {
              var args;
              args = 1 <= arguments.length ? __slice.call(arguments, 0) : [];
              return _this._runInnerDOMMethod.apply(_this, [method].concat(__slice.call(args)));
            }
          });
        };
      })(this));
    };

    InjectedComponent.prototype._verifyRequiredMethods = function() {
      var component;
      if (this.state.component != null) {
        component = this.state.component;
        return this.props.requiredMethods.forEach((function(_this) {
          return function(method) {
            var isMethodDefined;
            isMethodDefined = component.prototype[method] != null;
            if (!isMethodDefined) {
              throw new Error("" + component.name + " must implement method `" + method + "` when registering for " + (JSON.stringify(_this.props.matching)));
            }
          };
        })(this));
      }
    };

    InjectedComponent.prototype._getStateFromStores = function(props) {
      var component, components;
      if (props == null) {
        props = this.props;
      }
      components = ComponentRegistry.findComponentsMatching(props.matching);
      if (components.length > 1) {
        console.warn("There are multiple components available for " + (JSON.stringify(props.matching)) + ". <InjectedComponent> is only rendering the first one.");
      }
      component = components.length === 0 ? this.props.fallback : components[0];
      return {
        component: component,
        visible: ComponentRegistry.showComponentRegions()
      };
    };

    return InjectedComponent;

  })(React.Component);

  module.exports = InjectedComponent;

}).call(this);
