(function() {
  var AndCompositeMatcher, Matcher, NotCompositeMatcher, OrCompositeMatcher, SearchMatcher, doubleQuoteEscapeSequence, singleQuoteEscapeSequence, tableNameForJoin, _,
    __indexOf = [].indexOf || function(item) { for (var i = 0, l = this.length; i < l; i++) { if (i in this && this[i] === item) return i; } return -1; },
    __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; },
    __hasProp = {}.hasOwnProperty,
    __extends = function(child, parent) { for (var key in parent) { if (__hasProp.call(parent, key)) child[key] = parent[key]; } function ctor() { this.constructor = child; } ctor.prototype = parent.prototype; child.prototype = new ctor(); child.__super__ = parent.prototype; return child; };

  _ = require('underscore');

  tableNameForJoin = require('../models/utils').tableNameForJoin;

  singleQuoteEscapeSequence = "''";

  doubleQuoteEscapeSequence = '""';


  /*
  Public: The Matcher class encapsulates a particular comparison clause on an {Attribute}.
  Matchers can evaluate whether or not an object matches them, and also compose
  SQL clauses for the DatabaseStore. Each matcher has a reference to a model
  attribute, a comparator and a value.
  
  ```coffee
  
   * Retrieving Matchers
  
  isUnread = Thread.attributes.unread.equal(true)
  
  hasLabel = Thread.attributes.categories.contains('label-id-123')
  
   * Using Matchers in Database Queries
  
  DatabaseStore.findAll(Thread).where(isUnread)...
  
   * Using Matchers to test Models
  
  threadA = new Thread(unread: true)
  threadB = new Thread(unread: false)
  
  isUnread.evaluate(threadA)
   * => true
  isUnread.evaluate(threadB)
   * => false
  
  ```
  
  Section: Database
   */

  Matcher = (function() {
    function Matcher(attr, comparator, val) {
      this.attr = attr;
      this.comparator = comparator;
      this.val = val;
      this.muid = Matcher.muid;
      Matcher.muid = (Matcher.muid + 1) % 50;
      this;
    }

    Matcher.prototype.attribute = function() {
      return this.attr;
    };

    Matcher.prototype.value = function() {
      return this.val;
    };

    Matcher.prototype.evaluate = function(model) {
      var matcherValue, modelArrayContainsValue, modelValue;
      modelValue = model[this.attr.modelKey];
      if (modelValue instanceof Function) {
        modelValue = modelValue();
      }
      matcherValue = this.val;
      modelArrayContainsValue = function(array, searchItem) {
        var asId, item, search, _i, _len;
        asId = function(v) {
          if (v && v.id) {
            return v.id;
          } else {
            return v;
          }
        };
        search = asId(searchItem);
        for (_i = 0, _len = array.length; _i < _len; _i++) {
          item = array[_i];
          if (asId(item) === search) {
            return true;
          }
        }
        return false;
      };
      switch (this.comparator) {
        case '=':
          return modelValue === matcherValue;
        case '<':
          return modelValue < matcherValue;
        case '>':
          return modelValue > matcherValue;
        case '<=':
          return modelValue <= matcherValue;
        case '>=':
          return modelValue >= matcherValue;
        case 'in':
          return __indexOf.call(matcherValue, modelValue) >= 0;
        case 'contains':
          return modelArrayContainsValue(modelValue, matcherValue);
        case 'containsAny':
          return _.any(matcherValue, function(submatcherValue) {
            return modelArrayContainsValue(modelValue, submatcherValue);
          });
        case 'startsWith':
          return modelValue.startsWith(matcherValue);
        case 'like':
          return modelValue.search(new RegExp(".*" + matcherValue + ".*", "gi")) >= 0;
        default:
          throw new Error("Matcher.evaulate() not sure how to evaluate @{@attr.modelKey} with comparator " + this.comparator);
      }
    };

    Matcher.prototype.joinTableRef = function() {
      return "M" + this.muid;
    };

    Matcher.prototype.joinSQL = function(klass) {
      var joinTable, joinTableRef;
      switch (this.comparator) {
        case 'contains':
        case 'containsAny':
          joinTable = tableNameForJoin(klass, this.attr.itemClass);
          joinTableRef = this.joinTableRef();
          return "INNER JOIN `" + joinTable + "` AS `" + joinTableRef + "` ON `" + joinTableRef + "`.`id` = `" + klass.name + "`.`id`";
        default:
          return false;
      }
    };

    Matcher.prototype.whereSQL = function(klass) {
      var escaped, escapedVals, v, val, _i, _len;
      if (this.comparator === "like") {
        val = "%" + this.val + "%";
      } else {
        val = this.val;
      }
      if (_.isString(val)) {
        escaped = "'" + (val.replace(/'/g, singleQuoteEscapeSequence)) + "'";
      } else if (val === true) {
        escaped = 1;
      } else if (val === false) {
        escaped = 0;
      } else if (val instanceof Date) {
        escaped = val.getTime() / 1000;
      } else if (val instanceof Array) {
        escapedVals = [];
        for (_i = 0, _len = val.length; _i < _len; _i++) {
          v = val[_i];
          if (!_.isString(v)) {
            throw new Error("" + this.attr.jsonKey + " value " + v + " must be a string.");
          }
          escapedVals.push("'" + (v.replace(/'/g, singleQuoteEscapeSequence)) + "'");
        }
        escaped = "(" + (escapedVals.join(',')) + ")";
      } else {
        escaped = val;
      }
      switch (this.comparator) {
        case 'startsWith':
          return " RAISE `TODO`; ";
        case 'contains':
          return "`" + (this.joinTableRef()) + "`.`value` = " + escaped;
        case 'containsAny':
          return "`" + (this.joinTableRef()) + "`.`value` IN " + escaped;
        default:
          return "`" + klass.name + "`.`" + this.attr.jsonKey + "` " + this.comparator + " " + escaped;
      }
    };

    return Matcher;

  })();

  OrCompositeMatcher = (function(_super) {
    __extends(OrCompositeMatcher, _super);

    function OrCompositeMatcher(children) {
      this.children = children;
      this.whereSQL = __bind(this.whereSQL, this);
      this.joinSQL = __bind(this.joinSQL, this);
      this.evaluate = __bind(this.evaluate, this);
      this.value = __bind(this.value, this);
      this.attribute = __bind(this.attribute, this);
      this;
    }

    OrCompositeMatcher.prototype.attribute = function() {
      return null;
    };

    OrCompositeMatcher.prototype.value = function() {
      return null;
    };

    OrCompositeMatcher.prototype.evaluate = function(model) {
      var matcher, _i, _len, _ref;
      _ref = this.children;
      for (_i = 0, _len = _ref.length; _i < _len; _i++) {
        matcher = _ref[_i];
        if (matcher.evaluate(model)) {
          return true;
        }
      }
      return false;
    };

    OrCompositeMatcher.prototype.joinSQL = function(klass) {
      var join, joins, matcher, _i, _len, _ref;
      joins = [];
      _ref = this.children;
      for (_i = 0, _len = _ref.length; _i < _len; _i++) {
        matcher = _ref[_i];
        join = matcher.joinSQL(klass);
        if (join) {
          joins.push(join);
        }
      }
      if (joins.length) {
        return joins.join(" ");
      } else {
        return false;
      }
    };

    OrCompositeMatcher.prototype.whereSQL = function(klass) {
      var matcher, wheres, _i, _len, _ref;
      wheres = [];
      _ref = this.children;
      for (_i = 0, _len = _ref.length; _i < _len; _i++) {
        matcher = _ref[_i];
        wheres.push(matcher.whereSQL(klass));
      }
      return "(" + wheres.join(" OR ") + ")";
    };

    return OrCompositeMatcher;

  })(Matcher);

  AndCompositeMatcher = (function(_super) {
    __extends(AndCompositeMatcher, _super);

    function AndCompositeMatcher(children) {
      this.children = children;
      this.whereSQL = __bind(this.whereSQL, this);
      this.joinSQL = __bind(this.joinSQL, this);
      this.evaluate = __bind(this.evaluate, this);
      this.value = __bind(this.value, this);
      this.attribute = __bind(this.attribute, this);
      this;
    }

    AndCompositeMatcher.prototype.attribute = function() {
      return null;
    };

    AndCompositeMatcher.prototype.value = function() {
      return null;
    };

    AndCompositeMatcher.prototype.evaluate = function(model) {
      return _.every(this.children, function(matcher) {
        return matcher.evaluate(model);
      });
    };

    AndCompositeMatcher.prototype.joinSQL = function(klass) {
      var join, joins, matcher, _i, _len, _ref;
      joins = [];
      _ref = this.children;
      for (_i = 0, _len = _ref.length; _i < _len; _i++) {
        matcher = _ref[_i];
        join = matcher.joinSQL(klass);
        if (join) {
          joins.push(join);
        }
      }
      return joins;
    };

    AndCompositeMatcher.prototype.whereSQL = function(klass) {
      var matcher, wheres, _i, _len, _ref;
      wheres = [];
      _ref = this.children;
      for (_i = 0, _len = _ref.length; _i < _len; _i++) {
        matcher = _ref[_i];
        wheres.push(matcher.whereSQL(klass));
      }
      return "(" + wheres.join(" AND ") + ")";
    };

    return AndCompositeMatcher;

  })(Matcher);

  NotCompositeMatcher = (function(_super) {
    __extends(NotCompositeMatcher, _super);

    function NotCompositeMatcher(children) {
      this.children = children;
      this.whereSQL = __bind(this.whereSQL, this);
      this.joinSQL = __bind(this.joinSQL, this);
      this.evaluate = __bind(this.evaluate, this);
      this.value = __bind(this.value, this);
      this.attribute = __bind(this.attribute, this);
      this;
    }

    NotCompositeMatcher.prototype.attribute = function() {
      return null;
    };

    NotCompositeMatcher.prototype.value = function() {
      return null;
    };

    NotCompositeMatcher.prototype.evaluate = function(model) {
      return !_.every(this.children, function(matcher) {
        return matcher.evaluate(model);
      });
    };

    NotCompositeMatcher.prototype.joinSQL = function(klass) {
      var join, joins, matcher, _i, _len, _ref;
      joins = [];
      _ref = this.children;
      for (_i = 0, _len = _ref.length; _i < _len; _i++) {
        matcher = _ref[_i];
        join = matcher.joinSQL(klass);
        if (join) {
          joins.push(join);
        }
      }
      return joins;
    };

    NotCompositeMatcher.prototype.whereSQL = function(klass) {
      var matcher, wheres, _i, _len, _ref;
      wheres = [];
      _ref = this.children;
      for (_i = 0, _len = _ref.length; _i < _len; _i++) {
        matcher = _ref[_i];
        wheres.push(matcher.whereSQL(klass));
      }
      return "NOT (" + wheres.join(" AND ") + ")";
    };

    return NotCompositeMatcher;

  })(Matcher);

  SearchMatcher = (function(_super) {
    __extends(SearchMatcher, _super);

    function SearchMatcher(searchQuery) {
      this.whereSQL = __bind(this.whereSQL, this);
      this.joinSQL = __bind(this.joinSQL, this);
      this.evaluate = __bind(this.evaluate, this);
      this.value = __bind(this.value, this);
      this.attribute = __bind(this.attribute, this);
      SearchMatcher.__super__.constructor.call(this, null, null, null);
      this.searchQuery = searchQuery.trim().replace(/^['"]/, "").replace(/['"]$/, "").replace(/'/g, singleQuoteEscapeSequence).replace(/"/g, doubleQuoteEscapeSequence);
      this;
    }

    SearchMatcher.prototype.attribute = function() {
      return null;
    };

    SearchMatcher.prototype.value = function() {
      return null;
    };

    SearchMatcher.prototype.evaluate = function(model) {
      return true;
    };

    SearchMatcher.prototype.joinSQL = function(klass) {
      var joinTableRef, searchTable;
      searchTable = "" + klass.name + "Search";
      joinTableRef = this.joinTableRef();
      return "INNER JOIN `" + searchTable + "` AS `" + joinTableRef + "` ON `" + joinTableRef + "`.`content_id` = `" + klass.name + "`.`id`";
    };

    SearchMatcher.prototype.whereSQL = function(klass) {
      var searchTable;
      searchTable = "" + klass.name + "Search";
      return "`" + searchTable + "` MATCH '\"" + this.searchQuery + "\"'";
    };

    return SearchMatcher;

  })(Matcher);

  Matcher.muid = 0;

  Matcher.Or = OrCompositeMatcher;

  Matcher.And = AndCompositeMatcher;

  Matcher.Not = NotCompositeMatcher;

  Matcher.Search = SearchMatcher;

  module.exports = Matcher;

}).call(this);
