(function() {
  var AccountStore, Attributes, Contact, FocusedPerspectiveStore, Model, RegExpUtils, Utils, name_prefixes, name_suffixes, _,
    __hasProp = {}.hasOwnProperty,
    __extends = function(child, parent) { for (var key in parent) { if (__hasProp.call(parent, key)) child[key] = parent[key]; } function ctor() { this.constructor = child; } ctor.prototype = parent.prototype; child.prototype = new ctor(); child.__super__ = parent.prototype; return child; },
    __indexOf = [].indexOf || function(item) { for (var i = 0, l = this.length; i < l; i++) { if (i in this && this[i] === item) return i; } return -1; };

  Model = require('./model');

  Utils = require('./utils');

  Attributes = require('../attributes');

  RegExpUtils = require('../../regexp-utils');

  AccountStore = require('../stores/account-store');

  FocusedPerspectiveStore = null;

  _ = require('underscore');

  name_prefixes = {};

  name_suffixes = {};


  /*
  Public: The Contact model represents a Contact object served by the Nylas Platform API.
  For more information about Contacts on the Nylas Platform, read the
  [Contacts API Documentation](https://nylas.com/docs/api#contacts)
  
   *# Attributes
  
  `name`: {AttributeString} The name of the contact. Queryable.
  
  `email`: {AttributeString} The email address of the contact. Queryable.
  
  `thirdPartyData`: {AttributeObject} Extra data that we find out about a
  contact.  The data is keyed by the 3rd party service that dumped the data
  there. The value is an object of raw data in the form that the service
  provides
  
  We also have "normalized" optional data for each contact. This list may
  grow as the needs of a contact become more complex.
  
  This class also inherits attributes from {Model}
  
  Section: Models
   */

  Contact = (function(_super) {
    __extends(Contact, _super);

    function Contact() {
      if (this.thirdPartyData == null) {
        this.thirdPartyData = {};
      }
      Contact.__super__.constructor.apply(this, arguments);
    }

    Contact.attributes = _.extend({}, Model.attributes, {
      'name': Attributes.String({
        queryable: true,
        modelKey: 'name'
      }),
      'email': Attributes.String({
        queryable: true,
        modelKey: 'email'
      }),
      'thirdPartyData': Attributes.Object({
        modelKey: 'thirdPartyData'
      }),
      'title': Attributes.String({
        modelKey: 'title'
      }),
      'phone': Attributes.String({
        modelKey: 'phone'
      }),
      'company': Attributes.String({
        modelKey: 'company'
      })
    });

    Contact.additionalSQLiteConfig = {
      setup: function() {
        return ['CREATE INDEX IF NOT EXISTS ContactEmailIndex ON Contact(email)'];
      }
    };

    Contact.fromString = function(string, _arg) {
      var accountId, email, emailRegex, match, name, _ref;
      accountId = (_arg != null ? _arg : {}).accountId;
      emailRegex = RegExpUtils.emailRegex();
      match = emailRegex.exec(string);
      if (emailRegex.exec(string)) {
        throw new Error('Error while calling Contact.fromString: string contains more than one email');
      }
      email = match[0];
      name = string.slice(0, match.index - 1);
      if ((_ref = name[name.length - 1]) === '<' || _ref === '(') {
        name = name.slice(0, -1);
      }
      name = name.trim();
      return new Contact({
        accountId: accountId,
        name: name,
        email: email
      });
    };

    Contact.prototype.toString = function() {
      if (this.name && this.name !== this.email) {
        return "" + this.name + " <" + this.email + ">";
      } else {
        return this.email;
      }
    };

    Contact.prototype.toJSON = function() {
      var json;
      json = Contact.__super__.toJSON.apply(this, arguments);
      json['name'] || (json['name'] = json['email']);
      return json;
    };

    Contact.prototype.isValid = function() {
      var result;
      if (!this.email) {
        return false;
      }
      result = RegExpUtils.emailRegex().exec(this.email);
      if (result && result instanceof Array) {
        return result[0] === this.email;
      } else {
        return false;
      }
      return this.email.match(RegExpUtils.emailRegex()) !== null;
    };

    Contact.prototype.isMe = function() {
      var account;
      account = AccountStore.accountForEmail(this.email);
      return account != null;
    };

    Contact.prototype.isMePhrase = function(_arg) {
      var account, forceAccountLabel, includeAccountLabel, _ref;
      _ref = _arg != null ? _arg : {}, includeAccountLabel = _ref.includeAccountLabel, forceAccountLabel = _ref.forceAccountLabel;
      account = AccountStore.accountForEmail(this.email);
      if (!account) {
        return null;
      }
      if (includeAccountLabel) {
        if (FocusedPerspectiveStore == null) {
          FocusedPerspectiveStore = require('../stores/focused-perspective-store');
        }
        if (account && (FocusedPerspectiveStore.current().accountIds.length > 1 || forceAccountLabel)) {
          return "You (" + account.label + ")";
        }
      }
      return "You";
    };

    Contact.prototype.displayName = function(_arg) {
      var compact, forceAccountLabel, includeAccountLabel, _ref, _ref1, _ref2;
      _ref = _arg != null ? _arg : {}, includeAccountLabel = _ref.includeAccountLabel, compact = _ref.compact, forceAccountLabel = _ref.forceAccountLabel;
      if (compact == null) {
        compact = false;
      }
      if (includeAccountLabel == null) {
        includeAccountLabel = !compact;
      }
      if (compact) {
        return (_ref1 = this.isMePhrase({
          includeAccountLabel: includeAccountLabel,
          forceAccountLabel: forceAccountLabel
        })) != null ? _ref1 : this.firstName();
      } else {
        return (_ref2 = this.isMePhrase({
          includeAccountLabel: includeAccountLabel,
          forceAccountLabel: forceAccountLabel
        })) != null ? _ref2 : this.fullName();
      }
    };

    Contact.prototype.fullName = function() {
      return this._nameParts().join(' ');
    };

    Contact.prototype.firstName = function() {
      var articles, part, _i, _len, _ref, _ref1;
      articles = ['a', 'the'];
      _ref = this._nameParts();
      for (_i = 0, _len = _ref.length; _i < _len; _i++) {
        part = _ref[_i];
        if (_ref1 = part.toLowerCase(), __indexOf.call(articles, _ref1) < 0) {
          return part;
        }
      }
      return "";
    };

    Contact.prototype.lastName = function() {
      var _ref, _ref1;
      return (_ref = (_ref1 = this._nameParts().slice(1)) != null ? _ref1.join(" ") : void 0) != null ? _ref : "";
    };

    Contact.prototype.nameAbbreviation = function() {
      var c1, c2, _ref, _ref1, _ref2, _ref3;
      c1 = (_ref = (_ref1 = this.firstName()[0]) != null ? _ref1.toUpperCase() : void 0) != null ? _ref : "";
      c2 = (_ref2 = (_ref3 = this.lastName()[0]) != null ? _ref3.toUpperCase() : void 0) != null ? _ref2 : "";
      return c1 + c2;
    };

    Contact.prototype._nameParts = function() {
      var name, parts;
      name = this.name;
      if (!(name && name.length)) {
        name = this.email || "";
        if (name.indexOf('@') > 0) {
          name = name.split('@')[0];
        }
      }
      name = name.split(/(\svia\s)/i)[0];
      name = name.trim();
      parts = this._parseReverseNames(name);
      if (parts.join('').length === 0) {
        parts = [];
        parts = name.split(/\s+/);
        parts = _.reject(parts, function(part) {
          part = part.toLowerCase().replace(/\./, '');
          return (part in name_prefixes) || (part in name_suffixes);
        });
      }
      if (parts.join('').length === 0) {
        parts = [name];
      }
      if (parts.join('').length === 0) {
        parts = [this.email];
      }
      return parts;
    };

    Contact.prototype._parseReverseNames = function(name) {
      var description, firstName, lastName, parts, _ref, _ref1;
      parts = [];
      _ref = name.split(','), lastName = _ref[0], firstName = _ref[1];
      if (firstName) {
        _ref1 = firstName.split('('), firstName = _ref1[0], description = _ref1[1];
        parts.push(firstName.trim());
        parts.push(lastName.trim());
        if (description) {
          parts.push("(" + description.trim());
        }
      }
      return parts;
    };

    return Contact;

  })(Model);

  module.exports = Contact;

  _.each(['2dlt', '2lt', '2nd lieutenant', 'adm', 'administrative', 'admiral', 'amb', 'ambassador', 'attorney', 'atty', 'baron', 'baroness', 'bishop', 'br', 'brig gen or bg', 'brigadier general', 'brnss', 'brother', 'capt', 'captain', 'chancellor', 'chaplain', 'chapln', 'chief petty officer', 'cmdr', 'cntss', 'coach', 'col', 'colonel', 'commander', 'corporal', 'count', 'countess', 'cpl', 'cpo', 'cpt', 'doctor', 'dr', 'dr and mrs', 'drs', 'duke', 'ens', 'ensign', 'estate of', 'father', 'father', 'fr', 'frau', 'friar', 'gen', 'general', 'gov', 'governor', 'hon', 'honorable', 'judge', 'justice', 'lieutenant', 'lieutenant colonel', 'lieutenant commander', 'lieutenant general', 'lieutenant junior grade', 'lord', 'lt', 'ltc', 'lt cmdr', 'lt col', 'lt gen', 'ltg', 'lt jg', 'm', 'madame', 'mademoiselle', 'maj', 'maj', 'master sergeant', 'master sgt', 'miss', 'miss', 'mlle', 'mme', 'monsieur', 'monsignor', 'monsignor', 'mr', 'mr', 'mr & dr', 'mr and dr', 'mr & mrs', 'mr and mrs', 'mrs & mr', 'mrs and mr', 'ms', 'ms', 'msgr', 'msgr', 'ofc', 'officer', 'president', 'princess', 'private', 'prof', 'prof & mrs', 'professor', 'pvt', 'rabbi', 'radm', 'rear admiral', 'rep', 'representative', 'rev', 'reverend', 'reverends', 'revs', 'right reverend', 'rtrev', 's sgt', 'sargent', 'sec', 'secretary', 'sen', 'senator', 'senor', 'senora', 'senorita', 'sergeant', 'sgt', 'sgt', 'sheikh', 'sir', 'sister', 'sister', 'sr', 'sra', 'srta', 'staff sergeant', 'superintendent', 'supt', 'the hon', 'the honorable', 'the venerable', 'treas', 'treasurer', 'trust', 'trustees of', 'vadm', 'vice admiral'], function(prefix) {
    return name_prefixes[prefix] = true;
  });

  _.each(['1', '2', '3', '4', '5', '6', '7', 'i', 'ii', 'iii', 'iv', 'v', 'vi', 'vii', 'viii', 'ix', '1st', '2nd', '3rd', '4th', '5th', '6th', '7th', 'cfx', 'cnd', 'cpa', 'csb', 'csc', 'csfn', 'csj', 'dc', 'dds', 'esq', 'esquire', 'first', 'fs', 'fsc', 'ihm', 'jd', 'jr', 'md', 'ocd', 'ofm', 'op', 'osa', 'osb', 'osf', 'phd', 'pm', 'rdc', 'ret', 'rsm', 'second', 'sj', 'sm', 'snd', 'sp', 'sr', 'ssj', 'us army', 'us army ret', 'usa', 'usa ret', 'usaf', 'usaf ret', 'usaf us air force', 'usmc us marine corp', 'usmcr us marine reserves', 'usn', 'usn ret', 'usn us navy', 'vm'], function(suffix) {
    return name_suffixes[suffix] = true;
  });

}).call(this);
