(function() {
  var Attributes, Model, Utils, _;

  _ = require('underscore');

  Utils = require('./utils');

  Attributes = require('../attributes');


  /*
  Public: A base class for API objects that provides abstract support for
  serialization and deserialization, matching by attributes, and ID-based equality.
  
   *# Attributes
  
  `id`: {AttributeString} The resolved canonical ID of the model used in the
  database and generally throughout the app. The id property is a custom
  getter that resolves to the serverId first, and then the clientId.
  
  `clientId`: {AttributeString} An ID created at object construction and
  persists throughout the lifetime of the object. This is extremely useful
  for optimistically creating objects (like drafts and categories) and
  having a constant reference to it. In all other cases, use the resolved
  `id` field.
  
  `serverId`: {AttributeServerId} The server ID of the model. In most cases,
  except optimistic creation, this will also be the canonical id of the
  object.
  
  `object`: {AttributeString} The model's type. This field is used by the JSON
   deserializer to create an instance of the correct class when inflating the object.
  
  `accountId`: {AttributeString} The string Account Id this model belongs to.
  
  Section: Models
   */

  Model = (function() {
    Object.defineProperty(Model.prototype, "id", {
      enumerable: false,
      get: function() {
        var _ref;
        return (_ref = this.serverId) != null ? _ref : this.clientId;
      },
      set: function() {
        throw new Error("You may not directly set the ID of an object. Set either the `clientId` or the `serverId` instead.");
      }
    });

    Model.attributes = {
      'id': Attributes.String({
        queryable: true,
        modelKey: 'id'
      }),
      'clientId': Attributes.String({
        queryable: true,
        modelKey: 'clientId',
        jsonKey: 'client_id'
      }),
      'serverId': Attributes.ServerId({
        modelKey: 'serverId',
        jsonKey: 'server_id'
      }),
      'object': Attributes.String({
        modelKey: 'object'
      }),
      'accountId': Attributes.ServerId({
        queryable: true,
        modelKey: 'accountId',
        jsonKey: 'account_id'
      })
    };

    Model.naturalSortOrder = function() {
      return null;
    };

    function Model(values) {
      var key, _i, _len, _ref;
      if (values == null) {
        values = {};
      }
      if (values["id"] && Utils.isTempId(values["id"])) {
        if (values["clientId"] == null) {
          values["clientId"] = values["id"];
        }
      } else {
        if (values["serverId"] == null) {
          values["serverId"] = values["id"];
        }
      }
      _ref = Object.keys(this.constructor.attributes);
      for (_i = 0, _len = _ref.length; _i < _len; _i++) {
        key = _ref[_i];
        if (key === 'id') {
          continue;
        }
        if (values[key] == null) {
          continue;
        }
        this[key] = values[key];
      }
      if (this.clientId == null) {
        this.clientId = Utils.generateTempId();
      }
      this;
    }

    Model.prototype.isSavedRemotely = function() {
      return this.serverId != null;
    };

    Model.prototype.clone = function() {
      return (new this.constructor).fromJSON(this.toJSON());
    };

    Model.prototype.attributes = function() {
      var attrs;
      attrs = _.clone(this.constructor.attributes);
      delete attrs["id"];
      return attrs;
    };

    Model.prototype.fromJSON = function(json) {
      var attr, attrValue, key, _i, _len, _ref;
      if (json["id"] && !Utils.isTempId(json["id"])) {
        this.serverId = json["id"];
      }
      _ref = Object.keys(this.constructor.attributes);
      for (_i = 0, _len = _ref.length; _i < _len; _i++) {
        key = _ref[_i];
        if (key === 'id') {
          continue;
        }
        attr = this.constructor.attributes[key];
        attrValue = json[attr.jsonKey];
        if (attrValue !== void 0) {
          this[key] = attr.fromJSON(attrValue);
        }
      }
      return this;
    };

    Model.prototype.toJSON = function(options) {
      var attr, attrValue, json, key, _i, _len, _ref;
      if (options == null) {
        options = {};
      }
      json = {};
      _ref = Object.keys(this.constructor.attributes);
      for (_i = 0, _len = _ref.length; _i < _len; _i++) {
        key = _ref[_i];
        if (key === 'id') {
          continue;
        }
        attr = this.constructor.attributes[key];
        attrValue = this[key];
        if (attrValue === void 0) {
          continue;
        }
        if (attr instanceof Attributes.AttributeJoinedData && options.joined === false) {
          continue;
        }
        json[attr.jsonKey] = attr.toJSON(attrValue);
      }
      json["id"] = this.id;
      return json;
    };

    Model.prototype.toString = function() {
      return JSON.stringify(this.toJSON());
    };

    Model.prototype.matches = function(criteria) {
      var matcher, _i, _len;
      if (!(criteria instanceof Array)) {
        return false;
      }
      for (_i = 0, _len = criteria.length; _i < _len; _i++) {
        matcher = criteria[_i];
        if (!matcher.evaluate(this)) {
          return false;
        }
      }
      return true;
    };

    return Model;

  })();

  module.exports = Model;

}).call(this);
