(function() {
  var QueryRange, QueryResultSet, _;

  _ = require('underscore');

  QueryRange = require('./query-range');


  /*
  Public: Instances of QueryResultSet hold a set of models retrieved
  from the database at a given offset.
  
  Complete vs Incomplete:
  
  QueryResultSet keeps an array of item ids and a lookup table of models.
  The lookup table may be incomplete if the QuerySubscription isn't finished
  preparing results. You can use `isComplete` to determine whether the set
  has every model.
  
  Offset vs Index:
  
  To avoid confusion, "index" refers to an item's position in an
  array, and "offset" refers to it's position in the query result set. For example,
  an item might be at index 20 in the _ids array, but at offset 120 in the result.
  
  Ids and clientIds:
  
  QueryResultSet calways returns object `ids` when asked for ids, but lookups
  for models by clientId work once models are loaded.
   */

  QueryResultSet = (function() {
    QueryResultSet.setByApplyingModels = function(set, models) {
      if (models instanceof Array) {
        throw new Error("setByApplyingModels: A hash of models is required.");
      }
      set = set.clone();
      set._modelsHash = models;
      set._idToIndexHash = null;
      return set;
    };

    function QueryResultSet(other) {
      var _ref, _ref1, _ref2, _ref3, _ref4;
      if (other == null) {
        other = {};
      }
      this._offset = (_ref = other._offset) != null ? _ref : null;
      this._query = (_ref1 = other._query) != null ? _ref1 : null;
      this._idToIndexHash = (_ref2 = other._idToIndexHash) != null ? _ref2 : null;
      this._modelsHash = Object.assign({}, (_ref3 = other._modelsHash) != null ? _ref3 : {});
      this._ids = [].concat((_ref4 = other._ids) != null ? _ref4 : []);
    }

    QueryResultSet.prototype.clone = function() {
      return new this.constructor({
        _ids: [].concat(this._ids),
        _modelsHash: _.extend({}, this._modelsHash),
        _idToIndexHash: _.extend({}, this._idToIndexHash),
        _query: this._query,
        _offset: this._offset
      });
    };

    QueryResultSet.prototype.isComplete = function() {
      return _.every(this._ids, (function(_this) {
        return function(id) {
          return _this._modelsHash[id];
        };
      })(this));
    };

    QueryResultSet.prototype.range = function() {
      return new QueryRange({
        offset: this._offset,
        limit: this._ids.length
      });
    };

    QueryResultSet.prototype.query = function() {
      return this._query;
    };

    QueryResultSet.prototype.count = function() {
      return this._ids.length;
    };

    QueryResultSet.prototype.empty = function() {
      return this.count() === 0;
    };

    QueryResultSet.prototype.ids = function() {
      return this._ids;
    };

    QueryResultSet.prototype.idAtOffset = function(offset) {
      return this._ids[offset - this._offset];
    };

    QueryResultSet.prototype.models = function() {
      return this._ids.map((function(_this) {
        return function(id) {
          return _this._modelsHash[id];
        };
      })(this));
    };

    QueryResultSet.prototype.modelCacheCount = function() {
      return Object.keys(this._modelsHash).length;
    };

    QueryResultSet.prototype.modelAtOffset = function(offset) {
      if (!_.isNumber(offset)) {
        throw new Error("QueryResultSet.modelAtOffset() takes a numeric index. Maybe you meant modelWithId()?");
      }
      return this._modelsHash[this._ids[offset - this._offset]];
    };

    QueryResultSet.prototype.modelWithId = function(id) {
      return this._modelsHash[id];
    };

    QueryResultSet.prototype.buildIdToIndexHash = function() {
      var id, idx, model, _i, _len, _ref, _results;
      this._idToIndexHash = {};
      _ref = this._ids;
      _results = [];
      for (idx = _i = 0, _len = _ref.length; _i < _len; idx = ++_i) {
        id = _ref[idx];
        this._idToIndexHash[id] = idx;
        model = this._modelsHash[id];
        if (model) {
          _results.push(this._idToIndexHash[model.clientId] = idx);
        } else {
          _results.push(void 0);
        }
      }
      return _results;
    };

    QueryResultSet.prototype.offsetOfId = function(id) {
      if (this._idToIndexHash === null) {
        this.buildIdToIndexHash();
      }
      if (this._idToIndexHash[id] !== void 0) {
        return this._idToIndexHash[id] + this._offset;
      } else {
        return -1;
      }
    };

    return QueryResultSet;

  })();

  module.exports = QueryResultSet;

}).call(this);
