(function() {
  var AttributeCollection, AttributeJoinedData, Matcher, ModelQuery, QueryRange, Utils, _, _ref;

  _ref = require('../attributes'), Matcher = _ref.Matcher, AttributeJoinedData = _ref.AttributeJoinedData, AttributeCollection = _ref.AttributeCollection;

  QueryRange = require('./query-range');

  Utils = require('./utils');

  _ = require('underscore');


  /*
  Public: ModelQuery exposes an ActiveRecord-style syntax for building database queries
  that return models and model counts. Model queries are returned from the factory methods
  {DatabaseStore::find}, {DatabaseStore::findBy}, {DatabaseStore::findAll}, and {DatabaseStore::count}, and are the primary interface for retrieving data
  from the app's local cache.
  
  ModelQuery does not allow you to modify the local cache. To create, update or
  delete items from the local cache, see {DatabaseStore::persistModel}
  and {DatabaseStore::unpersistModel}.
  
  **Simple Example:** Fetch a thread
  
  ```coffee
  query = DatabaseStore.find(Thread, '123a2sc1ef4131')
  query.then (thread) ->
     * thread or null
  ```
  
  **Advanced Example:** Fetch 50 threads in the inbox, in descending order
  
  ```coffee
  query = DatabaseStore.findAll(Thread)
  query.where([Thread.attributes.categories.contains('label-id')])
       .order([Thread.attributes.lastMessageReceivedTimestamp.descending()])
       .limit(100).offset(50)
       .then (threads) ->
     * array of threads
  ```
  
  Section: Database
   */

  ModelQuery = (function() {
    function ModelQuery(_klass, _database) {
      this._klass = _klass;
      this._database = _database;
      this._database || (this._database = require('../stores/database-store'));
      this._matchers = [];
      this._orders = [];
      this._distinct = false;
      this._range = QueryRange.infinite();
      this._returnOne = false;
      this._returnIds = false;
      this._includeJoinedData = [];
      this._count = false;
      this;
    }

    ModelQuery.prototype.clone = function() {
      var q;
      q = new ModelQuery(this._klass, this._database).where(this._matchers).order(this._orders);
      q._orders = [].concat(this._orders);
      q._includeJoinedData = [].concat(this._includeJoinedData);
      q._range = this._range.clone();
      q._distinct = this._distinct;
      q._returnOne = this._returnOne;
      q._returnIds = this._returnIds;
      q._count = this._count;
      return q;
    };

    ModelQuery.prototype.distinct = function() {
      this._distinct = true;
      return this;
    };

    ModelQuery.prototype.where = function(matchers) {
      var attr, key, m, msg, value, _i, _len;
      this._assertNotFinalized();
      if (matchers instanceof Matcher) {
        this._matchers.push(matchers);
      } else if (matchers instanceof Array) {
        for (_i = 0, _len = matchers.length; _i < _len; _i++) {
          m = matchers[_i];
          if (!(m instanceof Matcher)) {
            throw new Error("You must provide instances of `Matcher`");
          }
        }
        this._matchers = this._matchers.concat(matchers);
      } else if (matchers instanceof Object) {
        for (key in matchers) {
          value = matchers[key];
          attr = this._klass.attributes[key];
          if (!attr) {
            msg = "Cannot create where clause `" + key + ":" + value + "`. " + key + " is not an attribute of " + this._klass.name;
            throw new Error(msg);
          }
          if (value instanceof Array) {
            this._matchers.push(attr["in"](value));
          } else {
            this._matchers.push(attr.equal(value));
          }
        }
      }
      return this;
    };

    ModelQuery.prototype.whereAny = function(matchers) {
      this._assertNotFinalized();
      this._matchers.push(new Matcher.Or(matchers));
      return this;
    };

    ModelQuery.prototype.search = function(query) {
      this._assertNotFinalized();
      this._matchers.push(new Matcher.Search(query));
      return this;
    };

    ModelQuery.prototype.include = function(attr) {
      this._assertNotFinalized();
      if (attr instanceof AttributeJoinedData === false) {
        throw new Error("query.include() must be called with a joined data attribute");
      }
      this._includeJoinedData.push(attr);
      return this;
    };

    ModelQuery.prototype.includeAll = function() {
      var attr, key, _ref1;
      this._assertNotFinalized();
      _ref1 = this._klass.attributes;
      for (key in _ref1) {
        attr = _ref1[key];
        if (attr instanceof AttributeJoinedData) {
          this.include(attr);
        }
      }
      return this;
    };

    ModelQuery.prototype.order = function(orders) {
      this._assertNotFinalized();
      if (!(orders instanceof Array)) {
        orders = [orders];
      }
      this._orders = this._orders.concat(orders);
      return this;
    };

    ModelQuery.prototype.one = function() {
      this._assertNotFinalized();
      this._returnOne = true;
      return this;
    };

    ModelQuery.prototype.limit = function(limit) {
      this._assertNotFinalized();
      if (this._returnOne && limit > 1) {
        throw new Error("Cannot use limit > 2 with one()");
      }
      this._range = this._range.clone();
      this._range.limit = limit;
      return this;
    };

    ModelQuery.prototype.offset = function(offset) {
      this._assertNotFinalized();
      this._range = this._range.clone();
      this._range.offset = offset;
      return this;
    };

    ModelQuery.prototype.page = function(start, end, pageSize, pagePadding) {
      var roundToPage;
      if (pageSize == null) {
        pageSize = 50;
      }
      if (pagePadding == null) {
        pagePadding = 100;
      }
      roundToPage = function(n) {
        return Math.max(0, Math.floor(n / pageSize) * pageSize);
      };
      this.offset(roundToPage(start - pagePadding));
      this.limit(roundToPage((end - start) + pagePadding * 2));
      return this;
    };

    ModelQuery.prototype.count = function() {
      this._assertNotFinalized();
      this._count = true;
      return this;
    };

    ModelQuery.prototype.idsOnly = function() {
      this._returnIds = true;
      return this;
    };


    /*
    Query Execution
     */

    ModelQuery.prototype.then = function(next) {
      return this.run(this).then(next);
    };

    ModelQuery.prototype.run = function() {
      return this._database.run(this);
    };

    ModelQuery.prototype.inflateResult = function(result) {
      var jsonError, objects;
      if (!result) {
        return null;
      }
      if (this._count) {
        return result[0]['count'] / 1;
      } else if (this._returnIds) {
        return result.map(function(row) {
          return row['id'];
        });
      } else {
        try {
          objects = result.map((function(_this) {
            return function(row) {
              var attr, json, object, value, _i, _len, _ref1;
              json = JSON.parse(row['data'], Utils.registeredObjectReviver);
              object = (new _this._klass).fromJSON(json);
              _ref1 = _this._includeJoinedData;
              for (_i = 0, _len = _ref1.length; _i < _len; _i++) {
                attr = _ref1[_i];
                value = row[attr.jsonKey];
                if (value === AttributeJoinedData.NullPlaceholder) {
                  value = null;
                }
                object[attr.modelKey] = value;
              }
              return object;
            };
          })(this));
        } catch (_error) {
          jsonError = _error;
          throw new Error("Query could not parse the database result. Query: " + (this.sql()) + ", Error: " + (jsonError.toString()));
        }
        return objects;
      }
    };

    ModelQuery.prototype.formatResult = function(inflated) {
      if (this._returnOne) {
        return inflated[0];
      }
      if (this._count) {
        return inflated;
      }
      return [].concat(inflated);
    };

    ModelQuery.prototype.sql = function() {
      var allMatchers, distinct, joins, limit, order, result, subSql;
      this.finalize();
      if (this._count) {
        result = "COUNT(*) as count";
      } else if (this._returnIds) {
        result = "`" + this._klass.name + "`.`id`";
      } else {
        result = "`" + this._klass.name + "`.`data`";
        this._includeJoinedData.forEach((function(_this) {
          return function(attr) {
            return result += ", " + (attr.selectSQL(_this._klass)) + " ";
          };
        })(this));
      }
      order = this._count ? "" : this._orderClause();
      if (this._range.limit != null) {
        limit = "LIMIT " + this._range.limit;
      } else {
        limit = "";
      }
      if (this._range.offset != null) {
        limit += " OFFSET " + this._range.offset;
      }
      distinct = this._distinct ? ' DISTINCT' : '';
      allMatchers = this.matchersFlattened();
      joins = allMatchers.filter(function(matcher) {
        return matcher.attr instanceof AttributeCollection;
      });
      if (joins.length === 1 && this._canSubselectForJoin(joins[0], allMatchers)) {
        subSql = this._subselectSQL(joins[0], this._matchers, order, limit);
        return "SELECT" + distinct + " " + result + " FROM `" + this._klass.name + "` WHERE `id` IN (" + subSql + ") " + order;
      } else {
        return "SELECT" + distinct + " " + result + " FROM `" + this._klass.name + "` " + (this._whereClause()) + " " + order + " " + limit;
      }
    };

    ModelQuery.prototype._canSubselectForJoin = function(matcher, allMatchers) {
      var allMatchersOnJoinTable, allOrdersOnJoinTable, joinAttribute;
      joinAttribute = matcher.attribute();
      if (this._range.limit == null) {
        return;
      }
      allMatchersOnJoinTable = _.every(allMatchers, function(m) {
        return m === matcher || joinAttribute.joinQueryableBy.indexOf(m.attr.modelKey) !== -1 || m.attr.modelKey === 'id';
      });
      allOrdersOnJoinTable = _.every(this._orders, function(o) {
        return joinAttribute.joinQueryableBy.indexOf(o.attr.modelKey) !== -1;
      });
      return allMatchersOnJoinTable && allOrdersOnJoinTable;
    };

    ModelQuery.prototype._subselectSQL = function(returningMatcher, subselectMatchers, order, limit) {
      var innerSQL, returningAttribute, table, wheres;
      returningAttribute = returningMatcher.attribute();
      table = Utils.tableNameForJoin(this._klass, returningAttribute.itemClass);
      wheres = _.compact(subselectMatchers.map((function(_this) {
        return function(c) {
          return c.whereSQL(_this._klass);
        };
      })(this)));
      innerSQL = "SELECT `id` FROM `" + table + "` WHERE " + (wheres.join(' AND ')) + " " + order + " " + limit;
      innerSQL = innerSQL.replace(new RegExp("`" + this._klass.name + "`", 'g'), "`" + table + "`");
      innerSQL = innerSQL.replace(new RegExp("`" + (returningMatcher.joinTableRef()) + "`", 'g'), "`" + table + "`");
      return innerSQL;
    };

    ModelQuery.prototype._whereClause = function() {
      var joins, sql, wheres;
      joins = [];
      this._matchers.forEach((function(_this) {
        return function(c) {
          var join;
          join = c.joinSQL(_this._klass);
          if (join) {
            return joins.push(join);
          }
        };
      })(this));
      this._includeJoinedData.forEach((function(_this) {
        return function(attr) {
          var join;
          join = attr.includeSQL(_this._klass);
          if (join) {
            return joins.push(join);
          }
        };
      })(this));
      wheres = [];
      this._matchers.forEach((function(_this) {
        return function(c) {
          var where;
          where = c.whereSQL(_this._klass);
          if (where) {
            return wheres.push(where);
          }
        };
      })(this));
      sql = "";
      sql += joins.join(' ');
      if (wheres.length) {
        sql += " WHERE " + wheres.join(' AND ');
      }
      return sql;
    };

    ModelQuery.prototype._orderClause = function() {
      var sql;
      if (!this._orders.length) {
        return "";
      }
      sql = " ORDER BY ";
      this._orders.forEach((function(_this) {
        return function(sort) {
          return sql += sort.orderBySQL(_this._klass);
        };
      })(this));
      return sql;
    };

    ModelQuery.prototype.finalize = function() {
      var natural;
      if (this._finalized) {
        return;
      }
      if (this._orders.length === 0) {
        natural = this._klass.naturalSortOrder();
        if (natural) {
          this._orders.push(natural);
        }
      }
      if (this._returnOne && !this._range.limit) {
        this.limit(1);
      }
      this._finalized = true;
      return this;
    };

    ModelQuery.prototype._assertNotFinalized = function() {
      if (this._finalized) {
        throw new Error("ModelQuery: You cannot modify a query after calling `then` or `listen`");
      }
    };

    ModelQuery.prototype.matchers = function() {
      return this._matchers;
    };

    ModelQuery.prototype.matchersFlattened = function() {
      var all, traverse;
      all = [];
      traverse = function(matchers) {
        var m, _i, _len, _results;
        if (!(matchers instanceof Array)) {
          return;
        }
        _results = [];
        for (_i = 0, _len = matchers.length; _i < _len; _i++) {
          m = matchers[_i];
          if (m.children) {
            _results.push(traverse(m.children));
          } else {
            _results.push(all.push(m));
          }
        }
        return _results;
      };
      traverse(this._matchers);
      return all;
    };

    ModelQuery.prototype.matcherValueForModelKey = function(key) {
      var matcher;
      matcher = _.find(this._matchers, function(m) {
        return m.attr.modelKey = key;
      });
      return matcher != null ? matcher.val : void 0;
    };

    ModelQuery.prototype.range = function() {
      return this._range;
    };

    ModelQuery.prototype.orderSortDescriptors = function() {
      return this._orders;
    };

    ModelQuery.prototype.objectClass = function() {
      return this._klass.name;
    };

    return ModelQuery;

  })();

  module.exports = ModelQuery;

}).call(this);
