(function() {
  var EventEmitter, callbackName, flattenListenables, mapChildListenables, _, _str;

  _ = require('underscore');

  _str = require('underscore.string');

  EventEmitter = require('events').EventEmitter;

  callbackName = function(string) {
    return "on" + string.charAt(0).toUpperCase() + string.slice(1);
  };


  /**
   * Extract child listenables from a parent from their
   * children property and return them in a keyed Object
   *
   * @param {Object} listenable The parent listenable
   */

  mapChildListenables = function(listenable) {
    var childName, children, i;
    i = 0;
    children = {};
    childName = void 0;
    while (i < (listenable.children || []).length) {
      childName = listenable.children[i];
      if (listenable[childName]) {
        children[childName] = listenable[childName];
      }
      ++i;
    }
    return children;
  };


  /**
   * Make a flat dictionary of all listenables including their
   * possible children (recursively), concatenating names in camelCase.
   *
   * @param {Object} listenables The top-level listenables
   */

  flattenListenables = function(listenables) {
    var childKey, childListenable, childMap, children, flattened, key, listenable;
    flattened = {};
    for (key in listenables) {
      listenable = listenables[key];
      childMap = mapChildListenables(listenable);
      children = flattenListenables(childMap);
      flattened[key] = listenable;
      for (childKey in children) {
        childListenable = children[childKey];
        flattened[key + _str.capitalize(childKey)] = childListenable;
      }
    }
    return flattened;
  };

  module.exports = {
    Listener: {
      hasListener: function(listenable) {
        var i, j, listenables, listener;
        i = 0;
        j = void 0;
        listener = void 0;
        listenables = void 0;
        while (i < (this.subscriptions || []).length) {
          listenables = [].concat(this.subscriptions[i].listenable);
          j = 0;
          while (j < listenables.length) {
            listener = listenables[j];
            if (listener === listenable || listener.hasListener && listener.hasListener(listenable)) {
              return true;
            }
            j++;
          }
          ++i;
        }
        return false;
      },
      listenToMany: function(listenables) {
        var allListenables, cbname, key, localname;
        allListenables = flattenListenables(listenables);
        for (key in allListenables) {
          cbname = callbackName(key);
          localname = this[cbname] ? cbname : this[key] ? key : void 0;
          if (localname) {
            this.listenTo(allListenables[key], localname, this[cbname + 'Default'] || this[localname + 'Default'] || localname);
          }
        }
      },
      validateListening: function(listenable) {
        if (listenable === this) {
          return 'Listener is not able to listen to itself';
        }
        if (!_.isFunction(listenable.listen)) {
          return listenable + ' is missing a listen method';
        }
        if (listenable.hasListener && listenable.hasListener(this)) {
          return 'Listener cannot listen to this listenable because of circular loop';
        }
      },
      listenTo: function(listenable, callback, defaultCallback) {
        var desub, err, resolvedCallback, subs, subscriptionobj, unsubscriber;
        desub = void 0;
        unsubscriber = void 0;
        subscriptionobj = void 0;
        subs = this.subscriptions = this.subscriptions || [];
        err = this.validateListening(listenable);
        if (err) {
          throw err;
        }
        this.fetchInitialState(listenable, defaultCallback);
        resolvedCallback = this[callback] || callback;
        if (!resolvedCallback) {
          throw new Error("@listenTo called with undefined callback");
        }
        desub = listenable.listen(resolvedCallback, this);
        unsubscriber = function() {
          var index;
          index = subs.indexOf(subscriptionobj);
          if (index === -1) {
            throw new Error('Tried to remove listen already gone from subscriptions list!');
          }
          subs.splice(index, 1);
          desub();
        };
        subscriptionobj = {
          stop: unsubscriber,
          listenable: listenable
        };
        subs.push(subscriptionobj);
        return subscriptionobj;
      },
      stopListeningTo: function(listenable) {
        var i, sub, subs;
        sub = void 0;
        i = 0;
        subs = this.subscriptions || [];
        while (i < subs.length) {
          sub = subs[i];
          if (sub.listenable === listenable) {
            sub.stop();
            if (subs.indexOf(sub) !== -1) {
              throw new Error('Failed to remove listen from subscriptions list!');
            }
            return true;
          }
          i++;
        }
        return false;
      },
      stopListeningToAll: function() {
        var remaining, subs;
        remaining = void 0;
        subs = this.subscriptions || [];
        while (remaining = subs.length) {
          subs[0].stop();
          if (subs.length !== remaining - 1) {
            throw new Error('Failed to remove listen from subscriptions list!');
          }
        }
      },
      fetchInitialState: function(listenable, defaultCallback) {
        var data, me;
        defaultCallback = defaultCallback && this[defaultCallback] || defaultCallback;
        me = this;
        if (_.isFunction(defaultCallback) && _.isFunction(listenable.getInitialState)) {
          data = listenable.getInitialState();
          if (data && _.isFunction(data.then)) {
            data.then(function() {
              defaultCallback.apply(me, arguments);
            });
          } else {
            defaultCallback.call(this, data);
          }
        }
      }
    },
    Publisher: {
      setupEmitter: function() {
        if (this._emitter) {
          return;
        }
        if (this._emitter == null) {
          this._emitter = new EventEmitter();
        }
        return this._emitter.setMaxListeners(50);
      },
      listen: function(callback, bindContext) {
        var aborted, eventHandler;
        if (!callback) {
          throw new Error("@listen called with undefined callback");
        }
        this.setupEmitter();
        if (bindContext == null) {
          bindContext = this;
        }
        aborted = false;
        eventHandler = function(args) {
          if (aborted) {
            return;
          }
          return callback.apply(bindContext, args);
        };
        this._emitter.addListener('trigger', eventHandler);
        return (function(_this) {
          return function() {
            aborted = true;
            return _this._emitter.removeListener('trigger', eventHandler);
          };
        })(this);
      },
      trigger: function() {
        this.setupEmitter();
        return this._emitter.emit('trigger', arguments);
      }
    }
  };

}).call(this);
