(function() {
  var APIError, Account, AccountStore, Actions, CancelledErrorCode, DatabaseStore, IdentityStore, Message, NylasAPI, NylasAPIChangeLockTracker, NylasAPIRequest, NylasLongConnection, PermanentErrorCodes, PriorityUICoordinator, SampleTemporaryErrorCode, TimeoutErrorCodes, Utils, async, remote, shell, _, _ref,
    __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; };

  _ = require('underscore');

  _ref = require('electron'), remote = _ref.remote, shell = _ref.shell;

  NylasLongConnection = require('./nylas-long-connection')["default"];

  Utils = require('./models/utils');

  Account = require('./models/account')["default"];

  Message = require('./models/message')["default"];

  IdentityStore = require('./stores/identity-store')["default"];

  Actions = require('./actions');

  APIError = require('./errors').APIError;

  PriorityUICoordinator = require('../priority-ui-coordinator');

  DatabaseStore = require('./stores/database-store');

  async = require('async');

  TimeoutErrorCodes = [0, "ETIMEDOUT", "ESOCKETTIMEDOUT", "ECONNRESET", "ENETDOWN", "ENETUNREACH"];

  PermanentErrorCodes = [400, 401, 402, 403, 404, 405, 500, "ENOTFOUND", "ECONNREFUSED", "EHOSTDOWN", "EHOSTUNREACH"];

  CancelledErrorCode = [-123, "ECONNABORTED"];

  SampleTemporaryErrorCode = 504;

  AccountStore = null;

  NylasAPIRequest = null;

  NylasAPIChangeLockTracker = (function() {
    function NylasAPIChangeLockTracker() {
      this._locks = {};
    }

    NylasAPIChangeLockTracker.prototype.acceptRemoteChangesTo = function(klass, id) {
      var key;
      key = "" + klass.name + "-" + id;
      return this._locks[key] === void 0;
    };

    NylasAPIChangeLockTracker.prototype.increment = function(klass, id) {
      var key, _base;
      key = "" + klass.name + "-" + id;
      if ((_base = this._locks)[key] == null) {
        _base[key] = 0;
      }
      return this._locks[key] += 1;
    };

    NylasAPIChangeLockTracker.prototype.decrement = function(klass, id) {
      var key;
      key = "" + klass.name + "-" + id;
      if (this._locks[key] == null) {
        return;
      }
      this._locks[key] -= 1;
      if (this._locks[key] <= 0) {
        return delete this._locks[key];
      }
    };

    NylasAPIChangeLockTracker.prototype.print = function() {
      console.log("The following models are locked:");
      return console.log(this._locks);
    };

    return NylasAPIChangeLockTracker;

  })();

  NylasAPI = (function() {
    NylasAPI.prototype.TimeoutErrorCodes = TimeoutErrorCodes;

    NylasAPI.prototype.PermanentErrorCodes = PermanentErrorCodes;

    NylasAPI.prototype.CancelledErrorCode = CancelledErrorCode;

    NylasAPI.prototype.SampleTemporaryErrorCode = SampleTemporaryErrorCode;

    function NylasAPI() {
      this._onConfigChanged = __bind(this._onConfigChanged, this);
      this._lockTracker = new NylasAPIChangeLockTracker();
      NylasEnv.config.onDidChange('env', this._onConfigChanged);
      this._onConfigChanged();
    }

    NylasAPI.prototype._onConfigChanged = function() {
      var current, env, prev;
      prev = {
        AppID: this.AppID,
        APIRoot: this.APIRoot,
        APITokens: this.APITokens
      };
      if (NylasEnv.inSpecMode()) {
        this.pluginsSupported = true;
        env = "testing";
      } else {
        env = NylasEnv.config.get('env');
      }
      if (!env) {
        env = 'production';
        console.warn("NylasAPI: config file does not contain an environment value. Defaulting to `production`.");
      }
      if (env === 'production') {
        this.AppID = 'eco3rpsghu81xdc48t5qugwq7';
        this.APIRoot = 'https://api.nylas.com';
        this.pluginsSupported = true;
      } else if (env === 'staging' || env === 'development') {
        this.AppID = '54miogmnotxuo5st254trcmb9';
        this.APIRoot = 'https://api-staging.nylas.com';
        this.pluginsSupported = true;
      } else if (env === 'experimental') {
        this.AppID = 'c5dis00do2vki9ib6hngrjs18';
        this.APIRoot = 'https://api-staging-experimental.nylas.com';
        this.pluginsSupported = true;
      } else if (env === 'local') {
        this.AppID = NylasEnv.config.get('syncEngine.AppID') || 'n/a';
        this.APIRoot = 'http://localhost:5555';
      } else if (env === 'custom') {
        this.AppID = NylasEnv.config.get('syncEngine.AppID') || 'n/a';
        this.APIRoot = NylasEnv.config.get('syncEngine.APIRoot') || 'http://localhost:5555';
      }
      return current = {
        AppID: this.AppID,
        APIRoot: this.APIRoot,
        APITokens: this.APITokens
      };
    };

    NylasAPI.prototype.makeRequest = function(options) {
      var error, req, success;
      if (options == null) {
        options = {};
      }
      if (NylasEnv.getLoadSettings().isSpec) {
        return Promise.resolve();
      }
      if (NylasAPIRequest == null) {
        NylasAPIRequest = require('./nylas-api-request')["default"];
      }
      req = new NylasAPIRequest(this, options);
      success = (function(_this) {
        return function(body) {
          if (options.beforeProcessing) {
            body = options.beforeProcessing(body);
          }
          if (options.returnsModel) {
            _this._handleModelResponse(body).then(function(objects) {
              return Promise.resolve(body);
            });
          }
          return Promise.resolve(body);
        };
      })(this);
      error = (function(_this) {
        return function(err) {
          var auth, handlePromise, returnsModel, url, _ref1, _ref2;
          _ref1 = req.options, url = _ref1.url, auth = _ref1.auth, returnsModel = _ref1.returnsModel;
          handlePromise = Promise.resolve();
          if (err.response) {
            if (err.response.statusCode === 404 && returnsModel) {
              handlePromise = _this._handleModel404(url);
            }
            if ((_ref2 = err.response.statusCode) === 401 || _ref2 === 403) {
              handlePromise = _this._handleAuthenticationFailure(url, auth != null ? auth.user : void 0);
            }
            if (err.response.statusCode === 400) {
              NylasEnv.reportError(err);
            }
          }
          return handlePromise["finally"](function() {
            return Promise.reject(err);
          });
        };
      })(this);
      return req.run().then(success, error);
    };

    NylasAPI.prototype.longConnection = function(opts) {
      var connection;
      connection = new NylasLongConnection(this, opts.accountId, opts);
      connection.onResults(opts.onResults);
      return connection;
    };

    NylasAPI.prototype.startLongConnection = function(opts) {
      return this.longConnection(opts).start();
    };

    NylasAPI.prototype._handleModel404 = function(modelUrl) {
      var collection, components, klass, klassId, pathname, query, root, url, _ref1;
      url = require('url');
      _ref1 = url.parse(modelUrl, true), pathname = _ref1.pathname, query = _ref1.query;
      components = pathname.split('/');
      if (components.length === 3) {
        root = components[0], collection = components[1], klassId = components[2];
        klass = this._apiObjectToClassMap[collection.slice(0, -1)];
      }
      if (klass && klassId && klassId.length > 0) {
        if (!NylasEnv.inSpecMode()) {
          console.warn("Deleting " + klass.name + ":" + klassId + " due to API 404");
        }
        return DatabaseStore.inTransaction(function(t) {
          return t.find(klass, klassId).then(function(model) {
            if (!model) {
              return Promise.resolve();
            }
            return t.unpersistModel(model);
          });
        });
      } else {
        return Promise.resolve();
      }
    };

    NylasAPI.prototype._handleAuthenticationFailure = function(modelUrl, apiToken) {
      var account;
      if (!apiToken) {
        return Promise.resolve();
      }
      if (AccountStore == null) {
        AccountStore = require('./stores/account-store');
      }
      account = AccountStore.accounts().find(function(account) {
        return AccountStore.tokenForAccountId(account.id) === apiToken;
      });
      if (account) {
        Actions.updateAccount(account.id, {
          syncState: Account.SYNC_STATE_AUTH_FAILED
        });
      }
      return Promise.resolve();
    };

    NylasAPI.prototype._handleModelResponse = function(jsons) {
      var ids, klass, type, uniquedJSONs, unlockedJSONs;
      if (!jsons) {
        return Promise.reject(new Error("handleModelResponse with no JSON provided"));
      }
      if (!(jsons instanceof Array)) {
        jsons = [jsons];
      }
      if (jsons.length === 0) {
        return Promise.resolve([]);
      }
      type = jsons[0].object;
      klass = this._apiObjectToClassMap[type];
      if (!klass) {
        console.warn("NylasAPI::handleModelResponse: Received unknown API object type: " + type);
        return Promise.resolve([]);
      }
      uniquedJSONs = _.uniq(jsons, false, function(model) {
        return model.id;
      });
      if (uniquedJSONs.length < jsons.length) {
        console.warn("NylasAPI::handleModelResponse: called with non-unique object set. Maybe an API request returned the same object more than once?");
      }
      unlockedJSONs = _.filter(uniquedJSONs, (function(_this) {
        return function(json) {
          var _ref1;
          if (_this._lockTracker.acceptRemoteChangesTo(klass, json.id) === false) {
            if ((_ref1 = json._delta) != null) {
              _ref1.ignoredBecause = "Model is locked, possibly because it's already been deleted.";
            }
            return false;
          }
          return true;
        };
      })(this));
      if (unlockedJSONs.length === 0) {
        return Promise.resolve([]);
      }
      ids = _.pluck(unlockedJSONs, 'id');
      return DatabaseStore.findAll(klass).where(klass.attributes.id["in"](ids)).then(function(models) {
        var changedModels, existingModels, model, responseModels, _i, _len;
        existingModels = {};
        for (_i = 0, _len = models.length; _i < _len; _i++) {
          model = models[_i];
          existingModels[model.id] = model;
        }
        responseModels = [];
        changedModels = [];
        unlockedJSONs.forEach((function(_this) {
          return function(json) {
            var isAlreadySent, isSameOrNewerVersion, _ref1, _ref2;
            model = existingModels[json.id];
            isSameOrNewerVersion = model && (model.version != null) && (json.version != null) && model.version >= json.version;
            isAlreadySent = model && model.draft === false && json.draft === true;
            if (isSameOrNewerVersion) {
              if ((_ref1 = json._delta) != null) {
                _ref1.ignoredBecause = "JSON v" + json.version + " <= model v" + model.version;
              }
            } else if (isAlreadySent) {
              if ((_ref2 = json._delta) != null) {
                _ref2.ignoredBecause = "Model " + model.id + " is already sent!";
              }
            } else {
              if (model == null) {
                model = new klass();
              }
              model.fromJSON(json);
              changedModels.push(model);
            }
            return responseModels.push(model);
          };
        })(this));
        return DatabaseStore.inTransaction(function(t) {
          return t.persistModels(changedModels);
        }).then(function() {
          return Promise.resolve(responseModels);
        });
      });
    };

    NylasAPI.prototype._attachMetadataToResponse = function(jsons, metadataToAttach) {
      var obj, _i, _len, _results;
      if (!metadataToAttach) {
        return;
      }
      _results = [];
      for (_i = 0, _len = jsons.length; _i < _len; _i++) {
        obj = jsons[_i];
        if (metadataToAttach[obj.id]) {
          _results.push(obj.metadata = metadataToAttach[obj.id]);
        } else {
          _results.push(void 0);
        }
      }
      return _results;
    };

    NylasAPI.prototype._apiObjectToClassMap = {
      "file": require('./models/file'),
      "event": require('./models/event'),
      "label": require('./models/label'),
      "folder": require('./models/folder'),
      "thread": require('./models/thread')["default"],
      "draft": require('./models/message')["default"],
      "account": require('./models/account')["default"],
      "message": require('./models/message')["default"],
      "contact": require('./models/contact'),
      "calendar": require('./models/calendar')
    };

    NylasAPI.prototype.getThreads = function(accountId, params, requestOptions) {
      var requestSuccess;
      if (params == null) {
        params = {};
      }
      if (requestOptions == null) {
        requestOptions = {};
      }
      requestSuccess = requestOptions.success;
      requestOptions.success = (function(_this) {
        return function(json) {
          var messages, result, _i, _len;
          messages = [];
          for (_i = 0, _len = json.length; _i < _len; _i++) {
            result = json[_i];
            if (result.messages) {
              messages = messages.concat(result.messages);
            }
          }
          if (messages.length > 0) {
            _this._attachMetadataToResponse(messages, requestOptions.metadataToAttach);
            _this._handleModelResponse(messages);
          }
          if (requestSuccess) {
            return requestSuccess(json);
          }
        };
      })(this);
      params.view = 'expanded';
      return this.getCollection(accountId, 'threads', params, requestOptions);
    };

    NylasAPI.prototype.getCollection = function(accountId, collection, params, requestOptions) {
      var requestSuccess;
      if (params == null) {
        params = {};
      }
      if (requestOptions == null) {
        requestOptions = {};
      }
      if (!accountId) {
        throw new Error("getCollection requires accountId");
      }
      requestSuccess = requestOptions.success;
      return this.makeRequest(_.extend(requestOptions, {
        path: "/" + collection,
        accountId: accountId,
        qs: params,
        returnsModel: false,
        success: (function(_this) {
          return function(jsons) {
            _this._attachMetadataToResponse(jsons, requestOptions.metadataToAttach);
            _this._handleModelResponse(jsons);
            if (requestSuccess) {
              return requestSuccess(jsons);
            }
          };
        })(this)
      }));
    };

    NylasAPI.prototype.makeDraftDeletionRequest = function(draft) {
      if (!draft.serverId) {
        return;
      }
      this.incrementRemoteChangeLock(Message, draft.serverId);
      this.makeRequest({
        path: "/drafts/" + draft.serverId,
        accountId: draft.accountId,
        method: "DELETE",
        body: {
          version: draft.version
        },
        returnsModel: false
      });
    };

    NylasAPI.prototype.incrementRemoteChangeLock = function(klass, id) {
      return this._lockTracker.increment(klass, id);
    };

    NylasAPI.prototype.decrementRemoteChangeLock = function(klass, id) {
      return this._lockTracker.decrement(klass, id);
    };

    NylasAPI.prototype.accessTokenForAccountId = function(aid) {
      if (AccountStore == null) {
        AccountStore = require('./stores/account-store');
      }
      return AccountStore.tokenForAccountId(aid);
    };

    NylasAPI.prototype.authPlugin = function(pluginId, pluginName, accountOrId) {
      var account, cacheKey;
      if (!this.pluginsSupported) {
        return Promise.reject(new Error('Sorry, this feature is only available when N1 is running against the hosted version of the Nylas Sync Engine.'));
      }
      if (accountOrId instanceof Account) {
        account = accountOrId;
      } else {
        if (AccountStore == null) {
          AccountStore = require('./stores/account-store');
        }
        account = AccountStore.accountForId(accountOrId);
      }
      if (!account) {
        return Promise.reject(new Error('Invalid account'));
      }
      cacheKey = "plugins." + pluginId + ".lastAuth." + account.id;
      if (NylasEnv.config.get(cacheKey)) {
        return Promise.resolve();
      }
      return this.makeRequest({
        returnsModel: false,
        method: "GET",
        accountId: account.id,
        path: "/auth/plugin?client_id=" + pluginId
      }).then((function(_this) {
        return function(result) {
          if (result.authed) {
            NylasEnv.config.set(cacheKey, Date.now());
            return Promise.resolve();
          }
          return _this.makeRequest({
            returnsModel: false,
            method: "POST",
            accountId: account.id,
            path: "/auth/plugin",
            body: {
              client_id: pluginId
            },
            json: true
          }).then(function() {
            NylasEnv.config.set(cacheKey, Date.now());
            return Promise.resolve();
          });
        };
      })(this));
    };

    NylasAPI.prototype._requestPluginAuth = function(pluginName, account) {
      return new Promise((function(_this) {
        return function(resolve, reject) {
          return remote.dialog.showMessageBox({
            title: "Plugin Offline Email Access",
            message: "The N1 plugin " + pluginName + " requests offline access to your email.",
            detail: "The " + pluginName + " plugin would like to be able to access your email account " + account.emailAddress + " while you are offline. Only grant offline access to plugins you trust. You can review and revoke Offline Access for plugins at any time from Preferences > Plugins.",
            buttons: ["Grant access", "Cancel"],
            type: 'info'
          }, function(result) {
            if (result === 0) {
              return resolve();
            } else {
              return reject();
            }
          });
        };
      })(this));
    };

    NylasAPI.prototype.unauthPlugin = function(pluginId, accountId) {
      return this.makeRequest({
        returnsModel: false,
        method: "DELETE",
        accountId: accountId,
        path: "/auth/plugin?client_id=" + pluginId
      });
    };

    return NylasAPI;

  })();

  module.exports = new NylasAPI();

}).call(this);
