(function() {
  var AccountStore, Actions, CoffeeHelpers, ContactStore, DatabaseStore, DraftChangeSet, DraftEditingSession, DraftHelpers, DraftStore, ExtensionRegistry, Listener, Message, MetadataChangePrefix, NylasAPI, Publisher, SyncbackDraftTask, UndoStack, _, _ref,
    __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; };

  Message = require('../models/message')["default"];

  Actions = require('../actions');

  NylasAPI = require('../nylas-api');

  AccountStore = require('./account-store');

  ContactStore = require('./contact-store');

  DatabaseStore = require('./database-store');

  UndoStack = require('../../undo-stack')["default"];

  DraftHelpers = require('../stores/draft-helpers');

  ExtensionRegistry = require('../../extension-registry');

  _ref = require('../modules/reflux-coffee'), Listener = _ref.Listener, Publisher = _ref.Publisher;

  SyncbackDraftTask = require('../tasks/syncback-draft-task')["default"];

  CoffeeHelpers = require('../coffee-helpers');

  DraftStore = null;

  _ = require('underscore');

  MetadataChangePrefix = 'metadata.';


  /*
  Public: As the user interacts with the draft, changes are accumulated in the
  DraftChangeSet associated with the store session. The DraftChangeSet does two things:
  
  1. It debounces changes and calls Actions.saveDraft() at a reasonable interval.
  
  2. It exposes `applyToModel`, which allows you to optimistically apply changes
    to a draft object. When the session vends the draft, it passes it through this
    function to apply uncommitted changes. This means the Draft provided by the
    DraftEditingSession will always relfect recent changes, even though they're
    written to the database intermittently.
  
  Section: Drafts
   */

  DraftChangeSet = (function() {
    DraftChangeSet.include = CoffeeHelpers.includeModule;

    DraftChangeSet.include(Publisher);

    function DraftChangeSet(callbacks) {
      this.callbacks = callbacks;
      this.applyToModel = __bind(this.applyToModel, this);
      this.commit = __bind(this.commit, this);
      this.addPluginMetadata = __bind(this.addPluginMetadata, this);
      this.add = __bind(this.add, this);
      this._commitChain = Promise.resolve();
      this._pending = {};
      this._saving = {};
      this._timer = null;
    }

    DraftChangeSet.prototype.teardown = function() {
      this._pending = {};
      this._saving = {};
      if (this._timer) {
        clearTimeout(this._timer);
        return this._timer = null;
      }
    };

    DraftChangeSet.prototype.add = function(changes, _arg) {
      var doesNotAffectPristine;
      doesNotAffectPristine = (_arg != null ? _arg : {}).doesNotAffectPristine;
      this.callbacks.onWillAddChanges(changes);
      this._pending = _.extend(this._pending, changes);
      if (!doesNotAffectPristine) {
        this._pending.pristine = false;
      }
      this.callbacks.onDidAddChanges(changes);
      if (this._timer) {
        clearTimeout(this._timer);
      }
      return this._timer = setTimeout(this.commit, 10000);
    };

    DraftChangeSet.prototype.addPluginMetadata = function(pluginId, metadata) {
      var changes;
      changes = {};
      changes["" + MetadataChangePrefix + pluginId] = metadata;
      return this.add(changes, {
        doesNotAffectPristine: true
      });
    };

    DraftChangeSet.prototype.commit = function(_arg) {
      var noSyncback;
      noSyncback = (_arg != null ? _arg : {}).noSyncback;
      if (this._timer) {
        clearTimeout(this._timer);
      }
      this._commitChain = this._commitChain["finally"]((function(_this) {
        return function() {
          if (Object.keys(_this._pending).length === 0) {
            return Promise.resolve(true);
          }
          _this._saving = _this._pending;
          _this._pending = {};
          return _this.callbacks.onCommit({
            noSyncback: noSyncback
          }).then(function() {
            return _this._saving = {};
          });
        };
      })(this));
      return this._commitChain;
    };

    DraftChangeSet.prototype.applyToModel = function(model) {
      var changesToApply, key, val, _i, _len, _ref1;
      if (model) {
        changesToApply = _.pairs(this._saving).concat(_.pairs(this._pending));
        for (_i = 0, _len = changesToApply.length; _i < _len; _i++) {
          _ref1 = changesToApply[_i], key = _ref1[0], val = _ref1[1];
          if (key.startsWith(MetadataChangePrefix)) {
            model.applyPluginMetadata(key.split(MetadataChangePrefix).pop(), val);
          } else {
            model[key] = val;
          }
        }
      }
      return model;
    };

    return DraftChangeSet;

  })();


  /*
  Public: DraftEditingSession is a small class that makes it easy to implement components
  that display Draft objects or allow for interactive editing of Drafts.
  
  1. It synchronously provides an instance of a draft via `draft()`, and
     triggers whenever that draft instance has changed.
  
  2. It provides an interface for modifying the draft that transparently
     batches changes, and ensures that the draft provided via `draft()`
     always has pending changes applied.
  
  Section: Drafts
   */

  DraftEditingSession = (function() {
    DraftEditingSession.include = CoffeeHelpers.includeModule;

    DraftEditingSession.include(Publisher);

    DraftEditingSession.include(Listener);

    function DraftEditingSession(draftClientId, draft) {
      this.draftClientId = draftClientId;
      if (draft == null) {
        draft = null;
      }
      this._snapshot = __bind(this._snapshot, this);
      this.redo = __bind(this.redo, this);
      this.undo = __bind(this.undo, this);
      this.restoreSnapshot = __bind(this.restoreSnapshot, this);
      this.changeSetDidAddChanges = __bind(this.changeSetDidAddChanges, this);
      this.changeSetWillAddChanges = __bind(this.changeSetWillAddChanges, this);
      this.changeSetCommit = __bind(this.changeSetCommit, this);
      this.ensureCorrectAccount = __bind(this.ensureCorrectAccount, this);
      this.validateDraftForSending = __bind(this.validateDraftForSending, this);
      if (DraftStore == null) {
        DraftStore = require('./draft-store');
      }
      this.listenTo(DraftStore, this._onDraftChanged);
      this._draft = false;
      this._draftPristineBody = null;
      this._destroyed = false;
      this._undoStack = new UndoStack();
      this.changes = new DraftChangeSet({
        onWillAddChanges: this.changeSetWillAddChanges,
        onDidAddChanges: this.changeSetDidAddChanges,
        onCommit: this.changeSetCommit
      });
      if (draft) {
        this._draftPromise = this._setDraft(draft);
      }
      this.prepare();
    }

    DraftEditingSession.prototype.draft = function() {
      if (!this._draft) {
        return null;
      }
      this.changes.applyToModel(this._draft);
      return this._draft.clone();
    };

    DraftEditingSession.prototype.draftPristineBody = function() {
      return this._draftPristineBody;
    };

    DraftEditingSession.prototype.prepare = function() {
      return this._draftPromise != null ? this._draftPromise : this._draftPromise = DatabaseStore.findBy(Message, {
        clientId: this.draftClientId
      }).include(Message.attributes.body).then((function(_this) {
        return function(draft) {
          if (_this._destroyed) {
            return Promise.reject(new Error("Draft has been destroyed."));
          }
          if (!draft) {
            return Promise.reject(new Error("Assertion Failure: Draft " + _this.draftClientId + " not found."));
          }
          return _this._setDraft(draft);
        };
      })(this));
    };

    DraftEditingSession.prototype.teardown = function() {
      this.stopListeningToAll();
      this.changes.teardown();
      return this._destroyed = true;
    };

    DraftEditingSession.prototype.validateDraftForSending = function() {
      var allRecipients, bodyIsEmpty, contact, errors, extension, forwarded, hasAttachment, warnings, _i, _j, _len, _len1, _ref1, _ref2, _ref3;
      warnings = [];
      errors = [];
      allRecipients = [].concat(this._draft.to, this._draft.cc, this._draft.bcc);
      bodyIsEmpty = this._draft.body === this.draftPristineBody();
      forwarded = DraftHelpers.isForwardedMessage(this._draft);
      hasAttachment = ((_ref1 = this._draft.files) != null ? _ref1.length : void 0) > 0 || ((_ref2 = this._draft.uploads) != null ? _ref2.length : void 0) > 0;
      for (_i = 0, _len = allRecipients.length; _i < _len; _i++) {
        contact = allRecipients[_i];
        if (!ContactStore.isValidContact(contact)) {
          errors.push("" + contact.email + " is not a valid email address - please remove or edit it before sending.");
        }
      }
      if (allRecipients.length === 0) {
        errors.push('You need to provide one or more recipients before sending the message.');
      }
      if (errors.length > 0) {
        return {
          errors: errors,
          warnings: warnings
        };
      }
      if (this._draft.subject.length === 0) {
        warnings.push('without a subject line');
      }
      if (DraftHelpers.messageMentionsAttachment(this._draft) && !hasAttachment) {
        warnings.push('without an attachment');
      }
      if (bodyIsEmpty && !forwarded && !hasAttachment) {
        warnings.push('without a body');
      }
      _ref3 = ExtensionRegistry.Composer.extensions();
      for (_j = 0, _len1 = _ref3.length; _j < _len1; _j++) {
        extension = _ref3[_j];
        if (!extension.warningsForSending) {
          continue;
        }
        warnings = warnings.concat(extension.warningsForSending({
          draft: this._draft
        }));
      }
      return {
        errors: errors,
        warnings: warnings
      };
    };

    DraftEditingSession.prototype.ensureCorrectAccount = function(_arg) {
      var account, noSyncback;
      noSyncback = (_arg != null ? _arg : {}).noSyncback;
      account = AccountStore.accountForEmail(this._draft.from[0].email);
      if (!account) {
        return Promise.reject(new Error("DraftEditingSession::ensureCorrectAccount - you can only send drafts from a configured account."));
      }
      if (account.id !== this._draft.accountId) {
        NylasAPI.makeDraftDeletionRequest(this._draft);
        this.changes.add({
          accountId: account.id,
          version: null,
          serverId: null,
          threadId: null,
          replyToMessageId: null
        });
        return this.changes.commit({
          noSyncback: noSyncback
        }).thenReturn(this);
      }
      return Promise.resolve(this);
    };

    DraftEditingSession.prototype._setDraft = function(draft) {
      if (draft.body == null) {
        throw new Error("DraftEditingSession._setDraft - new draft has no body!");
      }
      return Promise.each(ExtensionRegistry.Composer.extensions(), function(ext) {
        if (ext.applyTransformsToDraft && ext.unapplyTransformsToDraft) {
          return Promise.resolve(ext.unapplyTransformsToDraft({
            draft: draft
          })).then(function(untransformed) {
            if (untransformed !== 'unnecessary') {
              return draft = untransformed;
            }
          });
        }
      }).then((function(_this) {
        return function() {
          _this._draft = draft;
          if (draft.pristine) {
            _this._draftPristineBody = draft.body;
            _this._undoStack.save(_this._snapshot());
          }
          _this.trigger();
          return Promise.resolve(_this);
        };
      })(this));
    };

    DraftEditingSession.prototype._onDraftChanged = function(change) {
      var attr, key, nextDraft, nextValues, _ref1;
      if (change == null) {
        return;
      }
      if (!this._draft) {
        return;
      }
      nextDraft = _.filter(change.objects, (function(_this) {
        return function(obj) {
          return obj.clientId === _this._draft.clientId;
        };
      })(this)).pop();
      if (nextDraft) {
        nextValues = {};
        _ref1 = Message.attributes;
        for (key in _ref1) {
          attr = _ref1[key];
          if (key === 'id') {
            continue;
          }
          if (nextDraft[key] === void 0) {
            continue;
          }
          nextValues[key] = nextDraft[key];
        }
        this._setDraft(Object.assign(new Message(), this._draft, nextValues));
        return this.trigger();
      }
    };

    DraftEditingSession.prototype.changeSetCommit = function(_arg) {
      var inMemoryDraft, noSyncback;
      noSyncback = (_arg != null ? _arg : {}).noSyncback;
      if (this._destroyed || !this._draft) {
        return Promise.resolve(true);
      }
      inMemoryDraft = this._draft;
      return DatabaseStore.inTransaction((function(_this) {
        return function(t) {
          return t.findBy(Message, {
            clientId: inMemoryDraft.clientId
          }).include(Message.attributes.body).then(function(draft) {
            var updatedDraft;
            if (draft == null) {
              draft = inMemoryDraft;
            }
            updatedDraft = _this.changes.applyToModel(draft);
            return t.persistModel(updatedDraft);
          });
        };
      })(this)).then((function(_this) {
        return function() {
          if (noSyncback) {
            return;
          }
          if (!_this._draft.serverId) {
            return;
          }
          return Actions.ensureDraftSynced(_this.draftClientId);
        };
      })(this));
    };

    DraftEditingSession.prototype.changeSetWillAddChanges = function(changes) {
      var hasBeen300ms, hasChangedFields;
      if (this._restoring) {
        return;
      }
      hasBeen300ms = Date.now() - this._lastAddTimestamp > 300;
      hasChangedFields = !_.isEqual(Object.keys(changes), this._lastChangedFields);
      this._lastChangedFields = Object.keys(changes);
      this._lastAddTimestamp = Date.now();
      if (hasBeen300ms || hasChangedFields) {
        return this._undoStack.save(this._snapshot());
      }
    };

    DraftEditingSession.prototype.changeSetDidAddChanges = function() {
      if (this._destroyed) {
        return;
      }
      if (!this._draft) {
        throw new Error("DraftChangeSet was modified before the draft was prepared.");
      }
      this.changes.applyToModel(this._draft);
      return this.trigger();
    };

    DraftEditingSession.prototype.restoreSnapshot = function(snapshot) {
      if (!snapshot) {
        return;
      }
      this._restoring = true;
      this.changes.add(snapshot.draft);
      if (this._composerViewSelectionRestore) {
        this._composerViewSelectionRestore(snapshot.selection);
      }
      return this._restoring = false;
    };

    DraftEditingSession.prototype.undo = function() {
      return this.restoreSnapshot(this._undoStack.saveAndUndo(this._snapshot()));
    };

    DraftEditingSession.prototype.redo = function() {
      return this.restoreSnapshot(this._undoStack.redo());
    };

    DraftEditingSession.prototype._snapshot = function() {
      var pluginId, snapshot, value, _i, _len, _ref1, _ref2;
      snapshot = {
        selection: typeof this._composerViewSelectionRetrieve === "function" ? this._composerViewSelectionRetrieve() : void 0,
        draft: Object.assign({}, this.draft())
      };
      _ref1 = snapshot.draft.pluginMetadata;
      for (_i = 0, _len = _ref1.length; _i < _len; _i++) {
        _ref2 = _ref1[_i], pluginId = _ref2.pluginId, value = _ref2.value;
        snapshot.draft["" + MetadataChangePrefix + pluginId] = value;
      }
      delete snapshot.draft.pluginMetadata;
      return snapshot;
    };

    return DraftEditingSession;

  })();

  DraftEditingSession.DraftChangeSet = DraftChangeSet;

  module.exports = DraftEditingSession;

}).call(this);
