(function() {
  var AccountStore, Actions, BaseDraftTask, CoffeeHelpers, DatabaseStore, DestroyDraftTask, DraftEditingSession, DraftFactory, DraftHelpers, DraftStore, ExtensionRegistry, FocusedContentStore, Listener, Message, NylasAPI, Publisher, SendDraftTask, SoundRegistry, SyncbackDraftFilesTask, SyncbackDraftTask, TaskQueue, TaskQueueStatusStore, Thread, deprecate, ipcRenderer, store, _, _ref,
    __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; };

  _ = require('underscore');

  ipcRenderer = require('electron').ipcRenderer;

  NylasAPI = require('../nylas-api');

  DraftEditingSession = require('./draft-editing-session');

  DraftHelpers = require('./draft-helpers');

  DraftFactory = require('./draft-factory');

  DatabaseStore = require('./database-store');

  AccountStore = require('./account-store');

  TaskQueueStatusStore = require('./task-queue-status-store');

  FocusedContentStore = require('./focused-content-store');

  BaseDraftTask = require('../tasks/base-draft-task')["default"];

  SendDraftTask = require('../tasks/send-draft-task')["default"];

  SyncbackDraftFilesTask = require('../tasks/syncback-draft-files-task')["default"];

  SyncbackDraftTask = require('../tasks/syncback-draft-task')["default"];

  DestroyDraftTask = require('../tasks/destroy-draft-task')["default"];

  Thread = require('../models/thread')["default"];

  Message = require('../models/message')["default"];

  Actions = require('../actions');

  TaskQueue = require('./task-queue');

  SoundRegistry = require('../../sound-registry');

  _ref = require('../modules/reflux-coffee'), Listener = _ref.Listener, Publisher = _ref.Publisher;

  CoffeeHelpers = require('../coffee-helpers');

  ExtensionRegistry = require('../../extension-registry');

  deprecate = require('../../deprecate-utils').deprecate;


  /*
  Public: DraftStore responds to Actions that interact with Drafts and exposes
  public getter methods to return Draft objects and sessions.
  
  It also creates and queues {Task} objects to persist changes to the Nylas
  API.
  
  Remember that a "Draft" is actually just a "Message" with `draft: true`.
  
  Section: Drafts
   */

  DraftStore = (function() {
    DraftStore.include = CoffeeHelpers.includeModule;

    DraftStore.include(Publisher);

    DraftStore.include(Listener);

    function DraftStore() {
      this._onSendDraftSuccess = __bind(this._onSendDraftSuccess, this);
      this._onRemoveFile = __bind(this._onRemoveFile, this);
      this._onSendDraft = __bind(this._onSendDraft, this);
      this._onEnsureDraftSynced = __bind(this._onEnsureDraftSynced, this);
      this._onDestroyDraft = __bind(this._onDestroyDraft, this);
      this._onHandleMailFiles = __bind(this._onHandleMailFiles, this);
      this._onHandleMailtoLink = __bind(this._onHandleMailtoLink, this);
      this._onPopoutDraftClientId = __bind(this._onPopoutDraftClientId, this);
      this._onPopoutBlankDraft = __bind(this._onPopoutBlankDraft, this);
      this._createSession = __bind(this._createSession, this);
      this._finalizeAndPersistNewMessage = __bind(this._finalizeAndPersistNewMessage, this);
      this._onComposeForward = __bind(this._onComposeForward, this);
      this._onComposeReply = __bind(this._onComposeReply, this);
      this._onSendQuickReply = __bind(this._onSendQuickReply, this);
      this._onDataChanged = __bind(this._onDataChanged, this);
      this._onBeforeUnload = __bind(this._onBeforeUnload, this);
      this.unregisterExtension = __bind(this.unregisterExtension, this);
      this.registerExtension = __bind(this.registerExtension, this);
      this.extensions = __bind(this.extensions, this);
      this.sessionForClientId = __bind(this.sessionForClientId, this);
      this.listenTo(DatabaseStore, this._onDataChanged);
      this.listenTo(Actions.composeReply, this._onComposeReply);
      this.listenTo(Actions.composeForward, this._onComposeForward);
      this.listenTo(Actions.composePopoutDraft, this._onPopoutDraftClientId);
      this.listenTo(Actions.composeNewBlankDraft, this._onPopoutBlankDraft);
      this.listenTo(Actions.sendDraftFailed, this._onSendDraftFailed);
      this.listenTo(Actions.sendDraftSuccess, this._onSendDraftSuccess);
      this.listenTo(Actions.sendQuickReply, this._onSendQuickReply);
      if (NylasEnv.isMainWindow()) {
        ipcRenderer.on('new-message', (function(_this) {
          return function() {
            return Actions.composeNewBlankDraft();
          };
        })(this));
      }
      this.listenTo(Actions.ensureDraftSynced, this._onEnsureDraftSynced);
      this.listenTo(Actions.sendDraft, this._onSendDraft);
      this.listenTo(Actions.destroyDraft, this._onDestroyDraft);
      this.listenTo(Actions.removeFile, this._onRemoveFile);
      NylasEnv.onBeforeUnload(this._onBeforeUnload);
      this._draftSessions = {};
      this._draftsSending = {};
      ipcRenderer.on('mailto', this._onHandleMailtoLink);
      ipcRenderer.on('mailfiles', this._onHandleMailFiles);
    }

    DraftStore.prototype.sessionForClientId = function(clientId) {
      var _base;
      if (!clientId) {
        throw new Error("DraftStore::sessionForClientId requires a clientId");
      }
      if ((_base = this._draftSessions)[clientId] == null) {
        _base[clientId] = this._createSession(clientId);
      }
      return this._draftSessions[clientId].prepare();
    };

    DraftStore.prototype.isSendingDraft = function(draftClientId) {
      var _ref1;
      return (_ref1 = this._draftsSending[draftClientId]) != null ? _ref1 : false;
    };


    /*
    Composer Extensions
     */

    DraftStore.prototype.extensions = function() {
      return ExtensionRegistry.Composer.extensions();
    };

    DraftStore.prototype.registerExtension = function(ext) {
      return ExtensionRegistry.Composer.register(ext);
    };

    DraftStore.prototype.unregisterExtension = function(ext) {
      return ExtensionRegistry.Composer.unregister(ext);
    };

    DraftStore.prototype._doneWithSession = function(session) {
      session.teardown();
      return delete this._draftSessions[session.draftClientId];
    };

    DraftStore.prototype._cleanupAllSessions = function() {
      var draftClientId, session, _ref1, _results;
      _ref1 = this._draftSessions;
      _results = [];
      for (draftClientId in _ref1) {
        session = _ref1[draftClientId];
        _results.push(this._doneWithSession(session));
      }
      return _results;
    };

    DraftStore.prototype._onBeforeUnload = function(readyToUnload) {
      var key, promises, session, _ref1, _ref2;
      promises = [];
      _ref1 = this._draftSessions;
      for (key in _ref1) {
        session = _ref1[key];
        if ((_ref2 = session.draft()) != null ? _ref2.pristine : void 0) {
          Actions.queueTask(new DestroyDraftTask(session.draftClientId));
        } else {
          promises.push(session.changes.commit());
        }
      }
      if (promises.length > 0) {
        Promise.settle(promises).then((function(_this) {
          return function() {
            _this._draftSessions = {};
            return readyToUnload();
          };
        })(this));
        return false;
      } else {
        return true;
      }
    };

    DraftStore.prototype._onDataChanged = function(change) {
      var containsDraft;
      if (change.objectClass !== Message.name) {
        return;
      }
      containsDraft = _.some(change.objects, function(msg) {
        return msg.draft;
      });
      if (!containsDraft) {
        return;
      }
      return this.trigger(change);
    };

    DraftStore.prototype._onSendQuickReply = function(_arg, body) {
      var message, messageId, thread, threadId;
      thread = _arg.thread, threadId = _arg.threadId, message = _arg.message, messageId = _arg.messageId;
      return Promise.props(this._modelifyContext({
        thread: thread,
        threadId: threadId,
        message: message,
        messageId: messageId
      })).then((function(_this) {
        return function(_arg1) {
          var message, thread;
          message = _arg1.message, thread = _arg1.thread;
          return DraftFactory.createDraftForReply({
            message: message,
            thread: thread,
            type: 'reply'
          });
        };
      })(this)).then((function(_this) {
        return function(draft) {
          draft.body = body + "\n\n" + draft.body;
          draft.pristine = false;
          return DatabaseStore.inTransaction(function(t) {
            return t.persistModel(draft);
          }).then(function() {
            return Actions.sendDraft(draft.clientId);
          });
        };
      })(this));
    };

    DraftStore.prototype._onComposeReply = function(_arg) {
      var behavior, message, messageId, popout, thread, threadId, type;
      thread = _arg.thread, threadId = _arg.threadId, message = _arg.message, messageId = _arg.messageId, popout = _arg.popout, type = _arg.type, behavior = _arg.behavior;
      Actions.recordUserEvent("Draft Created", {
        type: type
      });
      return Promise.props(this._modelifyContext({
        thread: thread,
        threadId: threadId,
        message: message,
        messageId: messageId
      })).then((function(_this) {
        return function(_arg1) {
          var message, thread;
          message = _arg1.message, thread = _arg1.thread;
          return DraftFactory.createOrUpdateDraftForReply({
            message: message,
            thread: thread,
            type: type,
            behavior: behavior
          });
        };
      })(this)).then((function(_this) {
        return function(draft) {
          return _this._finalizeAndPersistNewMessage(draft, {
            popout: popout
          });
        };
      })(this));
    };

    DraftStore.prototype._onComposeForward = function(_arg) {
      var message, messageId, popout, thread, threadId;
      thread = _arg.thread, threadId = _arg.threadId, message = _arg.message, messageId = _arg.messageId, popout = _arg.popout;
      Actions.recordUserEvent("Draft Created", {
        type: "forward"
      });
      return Promise.props(this._modelifyContext({
        thread: thread,
        threadId: threadId,
        message: message,
        messageId: messageId
      })).then(DraftFactory.createDraftForForward).then((function(_this) {
        return function(draft) {
          return _this._finalizeAndPersistNewMessage(draft, {
            popout: popout
          });
        };
      })(this));
    };

    DraftStore.prototype._modelifyContext = function(_arg) {
      var message, messageId, queries, thread, threadId;
      thread = _arg.thread, threadId = _arg.threadId, message = _arg.message, messageId = _arg.messageId;
      queries = {};
      if (thread) {
        if (!(thread instanceof Thread)) {
          throw new Error("newMessageWithContext: `thread` present, expected a Model. Maybe you wanted to pass `threadId`?");
        }
        queries.thread = thread;
      } else {
        queries.thread = DatabaseStore.find(Thread, threadId);
      }
      if (message) {
        if (!(message instanceof Message)) {
          throw new Error("newMessageWithContext: `message` present, expected a Model. Maybe you wanted to pass `messageId`?");
        }
        queries.message = message;
      } else if (messageId != null) {
        queries.message = DatabaseStore.find(Message, messageId).include(Message.attributes.body);
      } else {
        queries.message = DatabaseStore.findBy(Message, {
          threadId: threadId != null ? threadId : thread.id
        }).order(Message.attributes.date.descending()).limit(1).include(Message.attributes.body);
      }
      return queries;
    };

    DraftStore.prototype._finalizeAndPersistNewMessage = function(draft, _arg) {
      var extension, popout, _i, _len, _ref1;
      popout = (_arg != null ? _arg : {}).popout;
      _ref1 = this.extensions();
      for (_i = 0, _len = _ref1.length; _i < _len; _i++) {
        extension = _ref1[_i];
        if (!extension.prepareNewDraft) {
          continue;
        }
        extension.prepareNewDraft({
          draft: draft
        });
      }
      this._createSession(draft.clientId, draft);
      return DatabaseStore.inTransaction((function(_this) {
        return function(t) {
          return t.persistModel(draft);
        };
      })(this)).then((function(_this) {
        return function() {
          if (popout) {
            return _this._onPopoutDraftClientId(draft.clientId);
          } else {
            return Actions.focusDraft({
              draftClientId: draft.clientId
            });
          }
        };
      })(this)).thenReturn({
        draftClientId: draft.clientId,
        draft: draft
      });
    };

    DraftStore.prototype._createSession = function(clientId, draft) {
      return this._draftSessions[clientId] = new DraftEditingSession(clientId, draft);
    };

    DraftStore.prototype._onPopoutBlankDraft = function() {
      Actions.recordUserEvent("Draft Created", {
        type: "new"
      });
      NylasEnv.perf.start("Popout Draft");
      return DraftFactory.createDraft().then((function(_this) {
        return function(draft) {
          return _this._finalizeAndPersistNewMessage(draft).then(function(_arg) {
            var draftClientId;
            draftClientId = _arg.draftClientId;
            return _this._onPopoutDraftClientId(draftClientId, {
              newDraft: true
            });
          });
        };
      })(this));
    };

    DraftStore.prototype._onPopoutDraftClientId = function(draftClientId, options) {
      var draftJSON, save, title;
      if (options == null) {
        options = {};
      }
      if (draftClientId == null) {
        throw new Error("DraftStore::onPopoutDraftId - You must provide a draftClientId");
      }
      NylasEnv.perf.start("Popout Draft");
      draftJSON = null;
      save = Promise.resolve();
      if (this._draftSessions[draftClientId]) {
        save = this._draftSessions[draftClientId].changes.commit();
        draftJSON = this._draftSessions[draftClientId].draft().toJSON();
      } else {
        save = this.sessionForClientId(draftClientId).then((function(_this) {
          return function(session) {
            return draftJSON = session.draft().toJSON();
          };
        })(this));
      }
      title = options.newDraft ? "New Message" : "Message";
      return save.then((function(_this) {
        return function() {
          return NylasEnv.newWindow({
            title: title,
            hidden: true,
            windowKey: "composer-" + draftClientId,
            windowType: "composer-preload",
            windowProps: _.extend(options, {
              draftClientId: draftClientId,
              draftJSON: draftJSON
            })
          });
        };
      })(this));
    };

    DraftStore.prototype._onHandleMailtoLink = function(event, urlString) {
      return DraftFactory.createDraftForMailto(urlString).then((function(_this) {
        return function(draft) {
          return _this._finalizeAndPersistNewMessage(draft, {
            popout: true
          });
        };
      })(this));
    };

    DraftStore.prototype._onHandleMailFiles = function(event, paths) {
      return DraftFactory.createDraft().then((function(_this) {
        return function(draft) {
          return _this._finalizeAndPersistNewMessage(draft);
        };
      })(this)).then((function(_this) {
        return function(_arg) {
          var callback, draftClientId, path, remaining, _i, _len, _results;
          draftClientId = _arg.draftClientId;
          remaining = paths.length;
          callback = function() {
            remaining -= 1;
            if (remaining === 0) {
              return _this._onPopoutDraftClientId(draftClientId);
            }
          };
          _results = [];
          for (_i = 0, _len = paths.length; _i < _len; _i++) {
            path = paths[_i];
            _results.push(Actions.addAttachment({
              filePath: path,
              messageClientId: draftClientId,
              callback: callback
            }));
          }
          return _results;
        };
      })(this));
    };

    DraftStore.prototype._onDestroyDraft = function(draftClientId) {
      var session, task, _i, _len, _ref1;
      session = this._draftSessions[draftClientId];
      if (session) {
        this._doneWithSession(session);
      }
      _ref1 = TaskQueueStatusStore.queue();
      for (_i = 0, _len = _ref1.length; _i < _len; _i++) {
        task = _ref1[_i];
        if (task instanceof BaseDraftTask && task.draftClientId === draftClientId) {
          Actions.dequeueTask(task.id);
        }
      }
      Actions.queueTask(new DestroyDraftTask(draftClientId));
      if (this._isPopout()) {
        return NylasEnv.close();
      }
    };

    DraftStore.prototype._onEnsureDraftSynced = function(draftClientId) {
      return this.sessionForClientId(draftClientId).then((function(_this) {
        return function(session) {
          return DraftHelpers.prepareDraftForSyncback(session).then(function(preparedDraft) {
            return Actions.queueTask(new SyncbackDraftTask(draftClientId));
          });
        };
      })(this));
    };

    DraftStore.prototype._onSendDraft = function(draftClientId) {
      this._draftsSending[draftClientId] = true;
      return this.sessionForClientId(draftClientId).then((function(_this) {
        return function(session) {
          return DraftHelpers.prepareDraftForSyncback(session).then(function(preparedDraft) {
            Actions.queueTask(new SendDraftTask(draftClientId));
            _this._doneWithSession(session);
            if (NylasEnv.config.get("core.sending.sounds")) {
              SoundRegistry.playSound('hit-send');
            }
            if (_this._isPopout()) {
              return NylasEnv.close();
            }
          });
        };
      })(this));
    };

    DraftStore.prototype.__testExtensionTransforms = function() {
      var clientId;
      clientId = NylasEnv.getWindowProps().draftClientId;
      return this.sessionForClientId(clientId).then((function(_this) {
        return function(session) {
          return _this._prepareForSyncback(session).then(function() {
            window.__draft = session.draft();
            return console.log("Done transforming draft. Available at window.__draft");
          });
        };
      })(this));
    };

    DraftStore.prototype._onRemoveFile = function(_arg) {
      var file, messageClientId;
      file = _arg.file, messageClientId = _arg.messageClientId;
      return this.sessionForClientId(messageClientId).then(function(session) {
        var files, _ref1;
        files = (_ref1 = _.clone(session.draft().files)) != null ? _ref1 : [];
        files = _.reject(files, function(f) {
          return f.id === file.id;
        });
        session.changes.add({
          files: files
        });
        return session.changes.commit();
      });
    };

    DraftStore.prototype._onSendDraftSuccess = function(_arg) {
      var draftClientId;
      draftClientId = _arg.draftClientId;
      delete this._draftsSending[draftClientId];
      return this.trigger(draftClientId);
    };

    DraftStore.prototype._onSendDraftFailed = function(_arg) {
      var draftClientId, errorMessage, threadId;
      draftClientId = _arg.draftClientId, threadId = _arg.threadId, errorMessage = _arg.errorMessage;
      this._draftsSending[draftClientId] = false;
      this.trigger(draftClientId);
      if (NylasEnv.isMainWindow()) {
        return _.delay((function(_this) {
          return function() {
            return _this._notifyUserOfError({
              draftClientId: draftClientId,
              threadId: threadId,
              errorMessage: errorMessage
            });
          };
        })(this), 300);
      }
    };

    DraftStore.prototype._isPopout = function() {
      return NylasEnv.getWindowType() === "composer";
    };

    DraftStore.prototype._notifyUserOfError = function(_arg) {
      var draftClientId, errorMessage, focusedThread, threadId;
      draftClientId = _arg.draftClientId, threadId = _arg.threadId, errorMessage = _arg.errorMessage;
      focusedThread = FocusedContentStore.focused('thread');
      if (threadId && (focusedThread != null ? focusedThread.id : void 0) === threadId) {
        return NylasEnv.showErrorDialog(errorMessage);
      } else {
        return Actions.composePopoutDraft(draftClientId, {
          errorMessage: errorMessage
        });
      }
    };

    return DraftStore;

  })();

  store = new DraftStore();

  store.registerExtension = deprecate('DraftStore.registerExtension', 'ExtensionRegistry.Composer.register', store, store.registerExtension);

  store.unregisterExtension = deprecate('DraftStore.unregisterExtension', 'ExtensionRegistry.Composer.unregister', store, store.unregisterExtension);

  module.exports = store;

}).call(this);
