(function() {
  var APIError, Actions, CoffeeHelpers, DatabaseStore, JSONBlobStorageKey, Listener, Publisher, Reflux, Task, TaskQueue, TaskRegistry, TimeoutError, Utils, fs, path, _, _ref, _ref1,
    __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; };

  _ = require('underscore');

  fs = require('fs-plus');

  path = require('path');

  _ref = require('../modules/reflux-coffee'), Listener = _ref.Listener, Publisher = _ref.Publisher;

  CoffeeHelpers = require('../coffee-helpers');

  Task = require("../tasks/task")["default"];

  TaskRegistry = require('../../task-registry')["default"];

  Utils = require("../models/utils");

  Reflux = require('reflux');

  Actions = require('../actions');

  DatabaseStore = require('./database-store');

  _ref1 = require('../errors'), APIError = _ref1.APIError, TimeoutError = _ref1.TimeoutError;

  JSONBlobStorageKey = 'task-queue';

  if (!NylasEnv.isWorkWindow() && !NylasEnv.inSpecMode()) {
    module.exports = {
      JSONBlobStorageKey: JSONBlobStorageKey
    };
    return;
  }


  /*
  Public: The TaskQueue is a Flux-compatible Store that manages a queue of {Task}
  objects. Each {Task} represents an individual API action, like sending a draft
  or marking a thread as "read". Tasks optimistically make changes to the app's
  local cache and encapsulate logic for performing changes on the server, rolling
  back in case of failure, and waiting on dependent tasks.
  
  The TaskQueue is essential to offline mode in N1. It automatically pauses
  when the user's internet connection is unavailable and resumes when online.
  
  The task queue is persisted to disk, ensuring that tasks are executed later,
  even if the user quits N1.
  
  The TaskQueue is only available in the app's main window. Rather than directly
  queuing tasks, you should use the {Actions} to interact with the {TaskQueue}.
  Tasks queued from secondary windows are serialized and sent to the application's
  main window via IPC.
  
   *# Queueing a Task
  
  ```coffee
  if @_thread && @_thread.unread
    Actions.queueTask(new ChangeStarredTask(thread: @_thread, starred: true))
  ```
  
   *# Dequeueing a Task
  
  ```coffee
  Actions.dequeueMatchingTask({
    type: 'DestroyCategoryTask',
    matching: {
      categoryId: 'bla'
    }
  })
  ```
  
   *# Creating Tasks
  
  Support for creating custom {Task} subclasses in third-party packages is coming soon.
  This is currently blocked by the ActionBridge, which is responsible for sending actions
  between windows, since it's JSON serializer is not extensible.
  
  Section: Stores
   */

  TaskQueue = (function() {
    TaskQueue.include = CoffeeHelpers.includeModule;

    TaskQueue.include(Publisher);

    TaskQueue.include(Listener);

    function TaskQueue() {
      this._ensurePeriodicUpdates = __bind(this._ensurePeriodicUpdates, this);
      this._updateSoon = __bind(this._updateSoon, this);
      this._restoreQueue = __bind(this._restoreQueue, this);
      this._resolveTaskArgument = __bind(this._resolveTaskArgument, this);
      this._taskIsBlocked = __bind(this._taskIsBlocked, this);
      this._dequeueObsoleteTasks = __bind(this._dequeueObsoleteTasks, this);
      this._processTask = __bind(this._processTask, this);
      this._processQueue = __bind(this._processQueue, this);
      this.clearCompleted = __bind(this.clearCompleted, this);
      this.dequeueMatching = __bind(this.dequeueMatching, this);
      this.dequeueAll = __bind(this.dequeueAll, this);
      this.dequeue = __bind(this.dequeue, this);
      this.enqueueUndoOfTaskId = __bind(this.enqueueUndoOfTaskId, this);
      this.enqueue = __bind(this.enqueue, this);
      this.allTasks = __bind(this.allTasks, this);
      this.completed = __bind(this.completed, this);
      this.queue = __bind(this.queue, this);
      this._queue = [];
      this._completed = [];
      this._updatePeriodicallyTimeout = null;
      this._currentSequentialId = Date.now();
      this._restoreQueue();
      this.listenTo(Actions.queueTask, this.enqueue);
      this.listenTo(Actions.queueTasks, (function(_this) {
        return function(tasks) {
          var t, _i, _len, _results;
          if (!(tasks && tasks.length > 0)) {
            return;
          }
          _results = [];
          for (_i = 0, _len = tasks.length; _i < _len; _i++) {
            t = tasks[_i];
            _results.push(_this.enqueue(t));
          }
          return _results;
        };
      })(this));
      this.listenTo(Actions.undoTaskId, this.enqueueUndoOfTaskId);
      this.listenTo(Actions.dequeueTask, this.dequeue);
      this.listenTo(Actions.dequeueAllTasks, this.dequeueAll);
      this.listenTo(Actions.dequeueMatchingTask, this.dequeueMatching);
      this.listenTo(Actions.clearDeveloperConsole, this.clearCompleted);
    }

    TaskQueue.prototype.queue = function() {
      return this._queue;
    };

    TaskQueue.prototype.completed = function() {
      return this._completed;
    };

    TaskQueue.prototype.allTasks = function() {
      return [].concat(this._queue, this._completed);
    };


    /*
    Public: Returns an existing task in the queue that matches the type you provide,
    and any other match properties. Useful for checking to see if something, like
    a "SendDraft" task is in-flight.
    
    - `type`: The string name of the task class, or the Task class itself. (ie:
      {SaveDraftTask} or 'SaveDraftTask')
    
    - `matching`: Optional An {Object} with criteria to pass to _.isMatch. For a
       SaveDraftTask, this could be {draftClientId: "123123"}
    
    Returns a matching {Task}, or null.
     */

    TaskQueue.prototype.findTask = function(type, matching) {
      if (matching == null) {
        matching = {};
      }
      return this.findTasks(type, matching)[0];
    };

    TaskQueue.prototype.findTasks = function(type, matching, _arg) {
      var includeCompleted, matches, tasks;
      if (matching == null) {
        matching = {};
      }
      includeCompleted = (_arg != null ? _arg : {}).includeCompleted;
      if (!_.isString(type)) {
        type = type.name;
      }
      tasks = includeCompleted ? this._queue.concat(this._completed) : this._queue;
      matches = _.filter(tasks, function(task) {
        var isMatch;
        if (task.constructor.name !== type) {
          return false;
        }
        isMatch = false;
        if (_.isFunction(matching)) {
          isMatch = matching(task);
        } else {
          isMatch = _.isMatch(task, matching);
        }
        return isMatch;
      });
      return matches != null ? matches : [];
    };

    TaskQueue.prototype.enqueue = function(task) {
      if (!(task instanceof Task)) {
        throw new Error("You must queue a `Task` instance");
      }
      if (!(TaskRegistry.isInRegistry(task.constructor.name))) {
        throw new Error("You must queue a `Task` instance which is registred with the TaskRegistry");
      }
      if (!task.id) {
        throw new Error("Tasks must have an ID prior to being queued. Check that your Task constructor is calling `super`");
      }
      if (!task.queueState) {
        throw new Error("Tasks must have a queueState prior to being queued. Check that your Task constructor is calling `super`");
      }
      task.sequentialId = ++this._currentSequentialId;
      this._dequeueObsoleteTasks(task);
      return task.runLocal().then((function(_this) {
        return function() {
          _this._queue.push(task);
          return _this._updateSoon();
        };
      })(this));
    };

    TaskQueue.prototype.enqueueUndoOfTaskId = function(taskId) {
      var task;
      task = _.findWhere(this._queue, {
        id: taskId
      });
      if (task == null) {
        task = _.findWhere(this._completed, {
          id: taskId
        });
      }
      if (task) {
        return this.enqueue(task.createUndoTask());
      }
    };

    TaskQueue.prototype.dequeue = function(taskOrId) {
      var task;
      task = this._resolveTaskArgument(taskOrId);
      if (!task) {
        throw new Error("Couldn't find task in queue to dequeue");
      }
      if (task.queueState.isProcessing) {
        return task.cancel();
      } else {
        this._queue.splice(this._queue.indexOf(task), 1);
        this._completed.push(task);
        if (this._completed.length > 1000) {
          this._completed.shift();
        }
        return this._updateSoon();
      }
    };

    TaskQueue.prototype.dequeueTaskAndDependents = function(taskOrId) {
      var task;
      task = this._resolveTaskArgument(taskOrId);
      if (!task) {
        throw new Error("Couldn't find task in queue to dequeue");
      }
    };

    TaskQueue.prototype.dequeueAll = function() {
      var task, _i, _ref2, _results;
      _ref2 = this._queue;
      _results = [];
      for (_i = _ref2.length - 1; _i >= 0; _i += -1) {
        task = _ref2[_i];
        _results.push(this.dequeue(task));
      }
      return _results;
    };

    TaskQueue.prototype.dequeueMatching = function(_arg) {
      var matching, task, type;
      type = _arg.type, matching = _arg.matching;
      task = this.findTask(type, matching);
      if (!task) {
        console.warn("Could not find matching task: " + type, matching);
        return;
      }
      return this.dequeue(task);
    };

    TaskQueue.prototype.clearCompleted = function() {
      this._completed = [];
      return this.trigger();
    };

    TaskQueue.prototype._processQueue = function() {
      var now, reprocessIn, started, task, _i, _ref2;
      started = 0;
      if (this._processQueueTimeout) {
        clearTimeout(this._processQueueTimeout);
        this._processQueueTimeout = null;
      }
      now = Date.now();
      reprocessIn = Number.MAX_VALUE;
      _ref2 = this._queue;
      for (_i = _ref2.length - 1; _i >= 0; _i += -1) {
        task = _ref2[_i];
        if (this._taskIsBlocked(task)) {
          task.queueState.debugStatus = Task.DebugStatus.WaitingOnDependency;
          continue;
        }
        if (task.queueState.retryAfter && task.queueState.retryAfter > now) {
          reprocessIn = Math.min(task.queueState.retryAfter - now, reprocessIn);
          task.queueState.debugStatus = Task.DebugStatus.WaitingToRetry;
          continue;
        }
        this._processTask(task);
        started += 1;
      }
      if (started > 0) {
        this.trigger();
      }
      if (reprocessIn !== Number.MAX_VALUE) {
        return this._processQueueTimeout = setTimeout(this._processQueue, reprocessIn + 500);
      }
    };

    TaskQueue.prototype._processTask = function(task) {
      if (task.queueState.isProcessing) {
        return;
      }
      task.queueState.isProcessing = true;
      return task.runRemote()["finally"]((function(_this) {
        return function() {
          task.queueState.isProcessing = false;
          return _this.trigger();
        };
      })(this)).then((function(_this) {
        return function(status) {
          var _ref2;
          if (status === Task.Status.Retry) {
            task.queueState.retryDelay = Math.round(Math.min(((_ref2 = task.queueState.retryDelay) != null ? _ref2 : 1000) * 2, 30000));
            task.queueState.retryAfter = Date.now() + task.queueState.retryDelay;
          } else {
            _this.dequeue(task);
          }
          return _this._updateSoon();
        };
      })(this))["catch"]((function(_this) {
        return function(err) {
          _this._seenDownstream = {};
          return _this._notifyOfDependentError(task, err).then(function(responses) {
            _this._dequeueDownstreamTasks(responses);
            return _this.dequeue(task);
          });
        };
      })(this));
    };

    TaskQueue.prototype._dequeueDownstreamTasks = function(responses) {
      if (responses == null) {
        responses = [];
      }
      responses = _.flatten(responses);
      responses = _.filter(responses, function(r) {
        return r != null;
      });
      return responses.forEach((function(_this) {
        return function(resp) {
          resp.downstreamTask.queueState.status = Task.Status.Continue;
          resp.downstreamTask.queueState.debugStatus = Task.DebugStatus.DequeuedDependency;
          return _this.dequeue(resp.downstreamTask);
        };
      })(this));
    };

    TaskQueue.prototype._notifyOfDependentError = function(failedTask, err) {
      var downstream, _ref2;
      downstream = (_ref2 = this._tasksToDequeueOnFailure(failedTask)) != null ? _ref2 : [];
      return Promise.map(downstream, (function(_this) {
        return function(downstreamTask) {
          var responseHash;
          if (!downstreamTask) {
            return Promise.resolve(null);
          }
          if (_this._seenDownstream[downstreamTask.id]) {
            return Promise.resolve(null);
          }
          _this._seenDownstream[downstreamTask.id] = true;
          responseHash = Promise.props({
            returnValue: downstreamTask.onDependentTaskError(failedTask, err),
            downstreamTask: downstreamTask
          });
          return Promise.all([responseHash, _this._notifyOfDependentError(downstreamTask, err)]);
        };
      })(this));
    };

    TaskQueue.prototype._dequeueObsoleteTasks = function(task) {
      var obsolete, otherTask, _i, _len, _results;
      obsolete = _.filter(this._queue, (function(_this) {
        return function(otherTask) {
          if (otherTask.queueState.isProcessing) {
            return false;
          }
          if (otherTask === task) {
            return false;
          }
          return task.shouldDequeueOtherTask(otherTask);
        };
      })(this));
      _results = [];
      for (_i = 0, _len = obsolete.length; _i < _len; _i++) {
        otherTask = obsolete[_i];
        otherTask.queueState.status = Task.Status.Continue;
        otherTask.queueState.debugStatus = Task.DebugStatus.DequeuedObsolete;
        _results.push(this.dequeue(otherTask));
      }
      return _results;
    };

    TaskQueue.prototype._tasksToDequeueOnFailure = function(failedTask) {
      return _.filter(this._queue, function(otherTask) {
        return failedTask !== otherTask && otherTask.isDependentOnTask(failedTask) && otherTask.shouldBeDequeuedOnDependencyFailure();
      });
    };

    TaskQueue.prototype._taskIsBlocked = function(task) {
      return _.any(this._queue, function(otherTask) {
        return task !== otherTask && task.isDependentOnTask(otherTask);
      });
    };

    TaskQueue.prototype._resolveTaskArgument = function(taskOrId) {
      if (!taskOrId) {
        return null;
      } else if (taskOrId instanceof Task) {
        return _.find(this._queue, function(task) {
          return task === taskOrId;
        });
      } else {
        return _.findWhere(this._queue, {
          id: taskOrId
        });
      }
    };

    TaskQueue.prototype._restoreQueue = function() {
      return DatabaseStore.findJSONBlob(JSONBlobStorageKey).then((function(_this) {
        return function(queue) {
          var task, _i, _len;
          if (queue == null) {
            queue = [];
          }
          for (_i = 0, _len = queue.length; _i < _len; _i++) {
            task = queue[_i];
            if (task.queueState == null) {
              task.queueState = {};
            }
            task.queueState.isProcessing = false;
            delete task.queueState['retryAfter'];
            delete task.queueState['retryDelay'];
          }
          queue = queue.filter(function(task) {
            return task instanceof Task;
          });
          _this._queue = queue;
          return _this._updateSoon();
        };
      })(this));
    };

    TaskQueue.prototype._updateSoon = function() {
      if (this._updateSoonThrottled == null) {
        this._updateSoonThrottled = _.throttle((function(_this) {
          return function() {
            DatabaseStore.inTransaction(function(t) {
              var _ref2;
              return t.persistJSONBlob(JSONBlobStorageKey, (_ref2 = _this._queue) != null ? _ref2 : []);
            });
            return _.defer(function() {
              _this._processQueue();
              return _this._ensurePeriodicUpdates();
            });
          };
        })(this), 10);
      }
      return this._updateSoonThrottled();
    };

    TaskQueue.prototype._ensurePeriodicUpdates = function() {
      var anyIsProcessing;
      anyIsProcessing = _.any(this._queue, function(task) {
        return task.queueState.isProcessing;
      });
      if (anyIsProcessing && !this._updatePeriodicallyTimeout) {
        return this._updatePeriodicallyTimeout = setInterval((function(_this) {
          return function() {
            return _this._updateSoon();
          };
        })(this), 1000);
      } else if (!anyIsProcessing && this._updatePeriodicallyTimeout) {
        clearTimeout(this._updatePeriodicallyTimeout);
        return this._updatePeriodicallyTimeout = null;
      }
    };

    return TaskQueue;

  })();

  module.exports = new TaskQueue();

  module.exports.JSONBlobStorageKey = JSONBlobStorageKey;

}).call(this);
