(function() {
  var AccountStore, Actions, Category, ChangeFolderTask, ChangeLabelsTask, ChangeStarredTask, ChangeUnreadTask, ConditionMode, ConditionTemplates, DatabaseStore, MailRulesActions, MailRulesProcessor, MailRulesStore, Message, Task, TaskQueueStatusStore, Thread, _, _ref,
    __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; };

  _ = require('underscore');

  Task = require('./flux/tasks/task')["default"];

  Actions = require('./flux/actions');

  Category = require('./flux/models/category');

  Thread = require('./flux/models/thread')["default"];

  Message = require('./flux/models/message')["default"];

  AccountStore = require('./flux/stores/account-store');

  DatabaseStore = require('./flux/stores/database-store');

  TaskQueueStatusStore = require('./flux/stores/task-queue-status-store');

  _ref = require('./mail-rules-templates'), ConditionMode = _ref.ConditionMode, ConditionTemplates = _ref.ConditionTemplates;

  ChangeUnreadTask = require('./flux/tasks/change-unread-task')["default"];

  ChangeFolderTask = require('./flux/tasks/change-folder-task')["default"];

  ChangeStarredTask = require('./flux/tasks/change-starred-task')["default"];

  ChangeLabelsTask = require('./flux/tasks/change-labels-task')["default"];

  MailRulesStore = null;


  /*
  Note: At first glance, it seems like these task factory methods should use the
  TaskFactory. Unfortunately, the TaskFactory uses the CategoryStore and other
  information about the current view. Maybe after the unified inbox refactor...
   */

  MailRulesActions = {
    markAsImportant: function(message, thread) {
      return DatabaseStore.findBy(Category, {
        name: 'important',
        accountId: thread.accountId
      }).then(function(important) {
        if (!important) {
          return Promise.reject(new Error("Could not find `important` label"));
        }
        return new ChangeLabelsTask({
          labelsToAdd: [important],
          threads: [thread.id]
        });
      });
    },
    moveToTrash: function(message, thread) {
      if (AccountStore.accountForId(thread.accountId).usesLabels()) {
        return MailRulesActions.moveToLabel(message, thread, 'trash');
      } else {
        return DatabaseStore.findBy(Category, {
          name: 'trash',
          accountId: thread.accountId
        }).then(function(folder) {
          if (!folder) {
            return Promise.reject(new Error("The folder could not be found."));
          }
          return new ChangeFolderTask({
            folder: folder,
            threads: [thread.id]
          });
        });
      }
    },
    markAsRead: function(message, thread) {
      return new ChangeUnreadTask({
        unread: false,
        threads: [thread.id]
      });
    },
    star: function(message, thread) {
      return new ChangeStarredTask({
        starred: true,
        threads: [thread.id]
      });
    },
    changeFolder: function(message, thread, value) {
      if (!value) {
        return Promise.reject(new Error("A folder is required."));
      }
      return DatabaseStore.findBy(Category, {
        id: value,
        accountId: thread.accountId
      }).then(function(folder) {
        if (!folder) {
          return Promise.reject(new Error("The folder could not be found."));
        }
        return new ChangeFolderTask({
          folder: folder,
          threads: [thread.id]
        });
      });
    },
    applyLabel: function(message, thread, value) {
      if (!value) {
        return Promise.reject(new Error("A label is required."));
      }
      return DatabaseStore.findBy(Category, {
        id: value,
        accountId: thread.accountId
      }).then(function(label) {
        if (!label) {
          return Promise.reject(new Error("The label could not be found."));
        }
        return new ChangeLabelsTask({
          labelsToAdd: [label],
          threads: [thread.id]
        });
      });
    },
    applyLabelArchive: function(message, thread) {
      return MailRulesActions.moveToLabel(message, thread, 'all');
    },
    moveToLabel: function(message, thread, nameOrId) {
      if (!nameOrId) {
        return Promise.reject(new Error("A label is required."));
      }
      return Promise.props({
        withId: DatabaseStore.findBy(Category, {
          id: nameOrId,
          accountId: thread.accountId
        }),
        withName: DatabaseStore.findBy(Category, {
          name: nameOrId,
          accountId: thread.accountId
        })
      }).then(function(_arg) {
        var label, withId, withName;
        withId = _arg.withId, withName = _arg.withName;
        label = withId || withName;
        if (!label) {
          return Promise.reject(new Error("The label could not be found."));
        }
        return new ChangeLabelsTask({
          labelsToRemove: [].concat(thread.labels).filter((function(_this) {
            return function(l) {
              return !l.isLockedCategory() && l.id !== label.id;
            };
          })(this)),
          labelsToAdd: [label],
          threads: [thread.id]
        });
      });
    }
  };

  MailRulesProcessor = (function() {
    function MailRulesProcessor() {
      this._applyRuleToMessage = __bind(this._applyRuleToMessage, this);
      this._checkRuleForMessage = __bind(this._checkRuleForMessage, this);
      this.processMessages = __bind(this.processMessages, this);
    }

    MailRulesProcessor.prototype.processMessages = function(messages) {
      var enabledRules;
      if (MailRulesStore == null) {
        MailRulesStore = require('./flux/stores/mail-rules-store');
      }
      if (!(messages.length > 0)) {
        return Promise.resolve();
      }
      enabledRules = MailRulesStore.rules().filter(function(r) {
        return !r.disabled;
      });
      return Promise.each(enabledRules, (function(_this) {
        return function(rule) {
          var matching;
          matching = messages.filter(function(message) {
            return _this._checkRuleForMessage(rule, message);
          });
          matching = _.uniq(matching, false, function(message) {
            return message.threadId;
          });
          return Promise.map(matching, function(message) {
            return DatabaseStore.find(Thread, message.threadId).then(function(thread) {
              if (!thread) {
                return console.warn("Cannot find thread " + message.threadId + " to process mail rules.");
              }
              return _this._applyRuleToMessage(rule, message, thread);
            });
          });
        };
      })(this));
    };

    MailRulesProcessor.prototype._checkRuleForMessage = function(rule, message) {
      var fn;
      if (rule.conditionMode === ConditionMode.All) {
        fn = _.every;
      } else {
        fn = _.any;
      }
      if (message.accountId !== rule.accountId) {
        return false;
      }
      return fn(rule.conditions, (function(_this) {
        return function(condition) {
          var template, value;
          template = _.findWhere(ConditionTemplates, {
            key: condition.templateKey
          });
          value = template.valueForMessage(message);
          return template.evaluate(condition, value);
        };
      })(this));
    };

    MailRulesProcessor.prototype._applyRuleToMessage = function(rule, message, thread) {
      var actionPromises;
      actionPromises = rule.actions.map((function(_this) {
        return function(action) {
          var actionFunction;
          actionFunction = MailRulesActions[action.templateKey];
          if (!actionFunction) {
            return Promise.reject(new Error("" + action.templateKey + " is not a supported action."));
          }
          return actionFunction(message, thread, action.value);
        };
      })(this));
      return Promise.all(actionPromises).then(function(actionResults) {
        var actionTasks, performLocalPromises;
        performLocalPromises = [];
        actionTasks = actionResults.filter(function(r) {
          return r instanceof Task;
        });
        actionTasks.forEach(function(task) {
          performLocalPromises.push(TaskQueueStatusStore.waitForPerformLocal(task));
          return Actions.queueTask(task);
        });
        return Promise.all(performLocalPromises);
      })["catch"](function(err) {
        Actions.disableMailRule(rule.id, err.toString());
        return Promise.resolve();
      });
    };

    return MailRulesProcessor;

  })();

  module.exports = new MailRulesProcessor;

}).call(this);
